# -*- coding: utf-8 -*-
#
# project-template documentation build configuration file, created by
# sphinx-quickstart on Mon Jan 18 14:44:12 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os

import sphinx_gallery
import sphinx_rtd_theme

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

sys.path.insert(0, os.path.abspath('documentation'))
sys.path.insert(0, os.path.abspath('kernels'))
sys.path.insert(0, os.path.abspath('generated'))
sys.path.insert(0, os.path.abspath('sphinxext'))
from github_link import make_linkcode_resolve

# -- General configuration ---------------------------------------------------

# Try to override the matplotlib configuration as early as possible
try:
    import gen_rst
except:
    pass
# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.linkcode',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.mathjax',
    'numpydoc',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx_gallery.gen_gallery',
    'sphinxcontrib.bibtex',
    'xref'
]

# Add external links
xref_links = {
    "lgscv": ("sklearn.model_selection.GridSearchCV",
              "http://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html"
              ),
    "gd": ("https://ls11-www.cs.tu-dortmund.de/staff/morris/graphkerneldatasets",
            "https://ls11-www.cs.tu-dortmund.de/staff/morris/graphkerneldatasets")
    }


doctest_global_setup = '''
import numpy as np
'''

# this is needed for some reason...
# see https://github.com/numpy/numpydoc/issues/69
numpydoc_class_members_toctree = False

sphinx_gallery_conf = {
    # path to your examples scripts
    'examples_dirs' : '../examples',

    # path where to save gallery generated examples
    'gallery_dirs'  : 'auto_examples',

    # for adding hyperlinks
    'reference_url': {'grakel': None},

    # backreferences for examples see
    # @https://github.com/sphinx-gallery/sphinx-gallery/blob/master/doc/advanced_configuration.rst#adding-references-to-examples
    'backreferences_dir' : os.path.join('generated'),
    
    # add define which module will be documented
    'doc_module': 'grakel',
}

autodoc_warningiserror=True

# Autodoc all members and inherited-members
autodoc_default_flags = ['members', 'inherited-members']

# Autoclass __init__
autoclass_content = 'both'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# generate autosummary even if no references
autosummary_generate = True

# The suffix of source filenames.
source_suffix = '.rst'


# The encoding of source files.
#source_encoding = 'utf-8-sig'

# Generate the plots for the gallery
plot_gallery = 'True'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'GraKeL'
copyright = u'2019, DaSciM (BSD License)'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
import grakel
version = grakel.__version__
# The full version, including alpha/beta/rc tags.
release = grakel.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
language = 'en'

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = False

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
#add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
#modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_figures/logo.svg"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Add any extra paths that contain custom files (such as robots.txt or
# .htaccess) here, relative to this directory. These files are copied
# directly to the root of the documentation.
#html_extra_path = []

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'grakel-doc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
# The paper size ('letterpaper' or 'a4paper').
#'papersize': 'letterpaper',

# The font size ('10pt', '11pt' or '12pt').
#'pointsize': '10pt',

# Additional stuff for the LaTeX preamble.
#'preamble': '',
}


# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
  ('index', 'grakel.tex', u'GraKeL Documentation',
   u'Yannis Siglidis', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'grakel', u'GraKeL Documentation',
     [u'Yannis Siglidis'], 1)
]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
  ('index', 'grakel', u'GraKeL Documentation',
   u'Yannis Siglidis', 'GraKeL', 'A library for graph kernels.',
   'Machine-Learning'),
]

def generate_example_rst(app, what, name, obj, options, lines):
    # generate empty examples files, so that we don't get
    # inclusion errors if there are no examples for a class / module
    examples_path = os.path.join(app.srcdir, "generated",
                                 "%s.examples" % name)
    if not os.path.exists(examples_path):
        # touch file
        open(examples_path, 'w').close()


def setup(app):
    app.add_css_file('_static/css/supplementary.css')
    app.connect('autodoc-process-docstring', generate_example_rst)

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'python': ('https://docs.python.org/{.major}'.format(
        sys.version_info), None),
    'numpy': ('https://docs.scipy.org/doc/numpy/', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
    'matplotlib': ('https://matplotlib.org/', None),
    'sklearn': ('http://scikit-learn.org/stable/', None),
    'cvxopt': ('http://cvxopt.org/', None),
    'pynauty': ('https://web.cs.dal.ca/~peter/software/pynauty/html/', None)
}

issues_uri = 'https://github.com/ysig/GraKeL/issues/{issue}'
issues_github_path = 'GraKeL/'
issues_user_uri = 'https://github.com/{user}'

# The following is used by sphinx.ext.linkcode to provide links to github
linkcode_resolve = make_linkcode_resolve('grakel',
                                         u'https://github.com/ysig/GraKeL'
                                         '/blob/{revision}/'
                                         '{package}/{path}#L{lineno}')



"""A file having general functions and classes usefull insid `grakel`."""

from __future__ import generators
import sys
import operator
import collections

import numpy as np

from scipy.special import binom as binomial

# Python 2/3 cross-compatibility import
from future.utils import iteritems
from six.moves.collections_abc import Hashable


class priority_dict(dict, object):
    """A priority dictionary.

    This implementation of priority dictionaries using binary heaps
    was implemented by David Eppstein, UC Irvine, 8 Mar 2002
    and was found at:
    http://code.activestate.com/recipes/117228-priority-dictionary/
    """

    def __init__(self):
        """Initialize a priority heap object."""
        self.__heap = []
        dict.__init__(self)

    def smallest(self):
        """Find smallest item after removing deleted items from heap."""
        if len(self) == 0:
            raise(IndexError, "smallest of empty priorityDictionary")
        heap = self.__heap
        while heap[0][1] not in self or self[heap[0][1]] != heap[0][0]:
            lastItem = heap.pop()
            insertionPoint = 0
            while 1:
                smallChild = 2*insertionPoint+1
                if smallChild+1 < len(heap) and \
                        heap[smallChild] > heap[smallChild+1]:
                    smallChild += 1
                if smallChild >= len(heap) or lastItem <= heap[smallChild]:
                    heap[insertionPoint] = lastItem
                    break
                heap[insertionPoint] = heap[smallChild]
                insertionPoint = smallChild
        return heap[0][1]

    def __iter__(self):
        """Create destructive sorted iterator of priorityDictionary."""
        def iterfn():
            while len(self) > 0:
                x = self.smallest()
                yield x
                del self[x]
        return iterfn()

    def __setitem__(self, key, val):
        """`__setitem__` default method for primary heap.

        Change value stored in dictionary and add corresponding
        pair to heap.  Rebuilds the heap if the number of deleted items grows
        too large, to avoid memory leakage.
        """
        dict.__setitem__(self, key, val)
        heap = self.__heap
        if len(heap) > 2 * len(self):
            self.__heap = [(v, k) for k, v in iteritems(self)]
            self.__heap.sort()  # builtin sort likely faster than O(n) heapify
        else:
            newPair = (val, key)
            insertionPoint = len(heap)
            heap.append(None)
            while insertionPoint > 0 and \
                    newPair < heap[(insertionPoint-1)//2]:
                heap[insertionPoint] = heap[(insertionPoint-1)//2]
                insertionPoint = (insertionPoint-1)//2
            heap[insertionPoint] = newPair

    def setdefault(self, key, val):
        """Reimplement setdefault to call our customized __setitem__."""
        if key not in self:
            self[key] = val
        return self[key]


def nested_dict_add(dictionary, value, *keys):
    """Nested adding in a dictionary.

    Parameters
    ----------
        dictionary : dict
            The dictionary that is going to be parsed.

        value : object
            The value that is going to be added.

        keys : list(hashable)
            A list of keys.
    Returns
    -------
    None.

    """
    address = dictionary
    for k in keys[:-1]:
        if(k not in address):
            address[k] = dict()
        address = address[k]
    address[keys[-1]] = value


def nested_dict_get(dictionary, *keys, **kargs):
    """Get an item from a nested dictionary.

    Checks if an item exists in a nested level by keys in a dictionary and if
    yes returns it. Otherwise return default.

    Parameters
    ----------
        dictionary : dict
            The dictionary that is going to be parsed.

        keys : list(hashable)
            A list of keys.

        default : object, default=None
            The value to return, if the element is not found.

    Returns
    -------
        dict_elem : object
            Returns either the dictionary element or the value in default.

    """
    if len(kargs) == 1 and "default":
        default = kargs["default"]
    elif len(kargs) == 0:
        default = None
    else:
        raise TypeError('optional argument can only be "default"')

    element = dictionary
    for k in keys:
        if (k in element):
            element = element[k]
        else:
            return default
    return element


def inv_dict(d):
    """Calculate the inverse dictionary.

    Parameters
    ----------
        d: dict
            A `surjective` dictionary.

    Returns
    -------
        inv_dict : dict
            The inverse dictionary.

    """
    inv = dict()
    if bool(d):
        for a in d.keys():
            k = d[a]
            if type(k) is list:
                k = tuple(k)
            elif type(k) is set:
                k = frozenset(k)
            else:
                if not isinstance(k, Hashable):
                    raise ValueError('in order to calculate inverse \
                          dictionary, values must be hashable')
            if k not in inv:
                inv[k] = list()
            inv[k].append(a)
        # returns a dictionary of lists
    return inv


ops = {
    '>': operator.gt,
    '<': operator.lt,
    '>=': operator.ge,
    '<=': operator.le,
    '==': operator.eq
    }


def matrix_to_dict(matrix, op='==', const_value=0, allow_diagonal=False):
    """Matrix to dictionary transformation based on constraints.

    Parameters
    ----------
        matrix : np.array
            A square to be transformed matrix.
        op : str
            An operator applied as a constraint.
        const_value : object
            The value to be compared with.
        allow_diagonal: bool
            Allows matrix diagonal to be added at input.

    Returns
    -------
        produced_dict : dict
            The produced matrix dictionary.

    """
    if op not in ['>', '<', '>=', '<=', '==']:
        raise ValueError('unsupported operator')

    opr = ops[op]

    s = matrix.shape[0]
    dictionary = dict()
    for i in range(0, s):
        line = matrix[i, :]
        if not allow_diagonal:
            np.delete(line, i)
        w = np.where(opr(line, const_value))
        dictionary[i] = list(w[0])
    return dictionary


def distribute_samples(n, subsets_size_range, n_samples):
    """Distribute samples evenly in a given range.

    A function that is used in order to distribute evenly, the amount of
    samples that will be drawn from a range of subset's sizes, from an
    original set of given size.

    Parameters
    ----------
        n : int
            The set size.

        subsets_size_range : tuple
            A touple having the min and the max subset size.

        n_samples : int
            The number of samples.

    Returns
    ------
        samples_on_subsets : dict
            Returns a dictionary of samples, for each subset.

    """
    # Check input
    min_ss, max_ss = subsets_size_range[0], subsets_size_range[1]

    # Distribute samples to subset groups
    maxd = min(max_ss, n)
    w = np.array([binomial(n, k) for k in range(min_ss, maxd+1)], dtype=float)
    w = w / np.sum(w)

    smpls = np.floor(w * n_samples).astype(int)
    ss = smpls.shape[0]

    for r in range(int(n_samples - np.sum(smpls))):
        smpls[(ss-r-1) % ss] += 1

    return {i + min_ss: smpls[i] for i in range(ss) if smpls[i] > .0}



"""A file that wraps c++ functions used in kernels submodule of grakel"""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause"
import numpy as np
import cython

from functools import reduce as freduce
from itertools import combinations
from collections import defaultdict

from numpy import floor, sqrt

cimport numpy as np

from libc.string cimport const_char
from libc.stdlib cimport malloc, free

from grakel.kernels._c_functions.header cimport ArashPartov, sm_core_init

def APHash(word):
    """C++ wrapped implementation of Arash Partov Hashing."""
    bs = word.encode('UTF-8')
    cdef int length = len(word);
    cdef const_char* string = bs;
    return ArashPartov(string, length)


def sm_kernel(x, y, kv, ke, k):
    """Calculate the weighted product graph and calculate the sm_kernel.

    For a definition of the weighted product graph see
    :cite:`Kriege2012SubgraphMK` (p.5, Definition 5).

    Parameters
    ----------
    x, y : tuples, size=4
        A tuple corresponding to the number of verices,
        the edge dictionarie starting from vertices of
        index zero, the labels for nodes, the labels for edges.

    kv : function
        A kernel for vertex labels.

    ke : function
        A kernel for edge labels.

    k : int
        The upper bound for the maximum size of subgraphs.

    Returns
    -------
    tv_np : np.array
        An array holding values for all clique sizes from 0 to k.

    """
    nx, Ex, Lx, Lex = x
    ny, Ey, Ly, Ley = y

    # Costs for vertices
    cv_l = list()

    if kv is not None:
        # Calculate valid vertices
        Vp = list()

        # calculate product graph vertex set
        nv = 0
        for i in range(nx):
            for j in range(ny):
                value = kv(Lx[i], Ly[j])
                if(value != .0):
                    # add to vertex set
                    Vp.append((i, j))
                    # initialise an empty set for neighbors
                    cv_l.append(value)
                    nv += 1
    else:
        Vp = [(i, j) for i in range(nx) for j in range(ny)]
        nv = nx*ny
        cv_l = iter(1. for _ in range(nv))

    # Initialise c arrays.
    cdef int *enum
    cdef double *cv
    cdef double **ce
    enum = <int *>malloc(nv*cython.sizeof(int))
    ce = <double **>malloc(nv*cython.sizeof(cython.p_double))
    cv = <double *>malloc(nv*cython.sizeof(cython.double))

    for (i, v) in enumerate(cv_l):
        enum[i] = i
        cv[i] = v
        ce[i] = <double *>malloc(nv*cython.sizeof(cython.double))

    with cython.boundscheck(False):
        # calculate product graph valid edges
        if ke is not None:
            for (i, v) in enumerate(Vp):
                for (j, w) in enumerate(Vp):
                    if i == j:
                        ce[j][i] = .0
                        break
                    if v[0] == w[0] or v[1] == w[1]:
                        value = .0
                    else:
                        ea, eb = (v[0], w[0]), (v[1], w[1])
                        conda, condb = ea not in Ex, eb not in Ey
                        if conda and condb:
                            # d-edge
                            value = -1.
                        elif conda or condb:
                            value = .0
                        else:
                            # possible c-edge
                            try:
                              lea = Lex[ea]
                              leb = Ley[eb]
                            except KeyError as key_error:
                              raise KeyError(key_error +
                                             '\nEdge labels must exist for '
                                             'all edges.')
                            value = ke(lea, leb)

                    ce[j][i] = ce[i][j] = value
        else:
            for (i, v) in enumerate(Vp):
                for (j, w) in enumerate(Vp):
                    if i == j:
                        ce[j][i] = .0
                        break
                    if v[0] == w[0] or v[1] == w[1]:
                        value = .0
                    else:
                        ea, eb = (v[0], w[0]), (v[1], w[1])
                        conda, condb = ea not in Ex, eb not in Ey
                        if conda and condb:
                            # d-edge
                            value = -1.
                        elif conda or condb:
                            value = .0
                        else:
                            value = 1.
                    ce[j][i] = ce[i][j] = value
            

    # Initialize values
    cdef np.ndarray[double, ndim=1] tv_np = np.zeros(shape=(k + 1))
    cdef double *tv = &tv_np[0]

    try:
        # Run the core function
        sm_core_init(1, enum, nv, k, cv, ce, tv)

        tv_np.reshape((k+1, 1))
        return tv_np
    finally:
        # Deallocate memory
        free(enum)
        free(cv)
        for i in range(nv):
            free(ce[i])
        free(ce)

def k_to_ij_triangular(k, dim):
    i = int(dim - 1 - floor(sqrt(-8*k + 4*(dim+1)*dim-7)/2.0 - 0.5))
    j = int(k + i - (dim+1)*(dim)//2 + (dim-i+1)*(dim-i)//2)
    return (i, j)

def k_to_ij_rectangular(k, dim):
    i = k % dim
    j = k // dim
    return (i, j)


# ConSubg from:
# Karakashian, Shant Kirakos et al. “An Algorithm for Generating All Connected Subgraphs with k Vertices of a Graph.” (2013).
def ConSubg(G, k, symmetric):
    # G: dict of sets
    l = set()
    if symmetric:
        sG = G
        for u in G.keys():
            l |= CombinationsWithV(u, k, sG)    
            sGP = dict()
            for v in sG.keys():
                if u != v:
                    sGP[v] = sG[v] - {u}
            sG = sGP
    else:
        for u in G.keys():
            l |= CombinationsWithV(u, k, G)

    return l

def CombinationsWithV(u, k, G_init):
    l = list()
    tree = defaultdict(set)
    treeL = {0: u}
    MarkN = dict()
    def CombinationTree(u, k, G):
        root = u
        l = [set() for i in range(k)]
        l[0].add(u)

        MarkV = dict()
        def BuildTree(nt, depth, k):
            # globals l, MarkN, MarkV, tree
            l[depth] = set(l[depth-1])
            for v in G[treeL[nt]]:
                if v != nt and v not in l[depth]:
                    ntp = len(treeL)
                    treeL[ntp] = v
                    tree[nt].add(ntp)
                    l[depth].add(v)
                    if not MarkV.get(v, False):
                        MarkN[ntp], MarkV[v] = True, True
                    else:
                        MarkN[ntp] = False
                    if depth + 1 <= k-1:
                        BuildTree(ntp, depth + 1, k)

        BuildTree(0, 1, k)

    def unionProduct(S1, S2):
        # globals tree, MarkN
        # print("To compare", S1, S2)
        if not len(S1):
            return set()
        elif not len(S2):
            return {S1}
        else:
            return {s1 | s2 for s1 in S1 for s2 in S2 for s1p, s2p in [({treeL[i] for i in s1}, {treeL[i] for i in s2})] if not len(s1p & {treeL[i] for i in s2}) and (any(MarkN[j] for j in s2) or all(not len({treeL[j] for j in tree[i]} & s2p) for i in s1))}
      
    # Memoization
    CFM = dict()

    def CombinationsFromTree(root, k):
        # Globals tree
        t = root
        lnodesets = set()
        if k == 1:
            return {frozenset({t})}
        for i in range(1, min(len(tree[t]), k - 1) + 1):
            for NodeComb in combinations(tree[t], i):
                for string in compositions(k - 1, i):
                    fail = False
                    S = list()
                    for pos in range(i):
                        stRoot = NodeComb[pos]
                        size = string[pos]
                        m = CFM.get((stRoot, size), None)
                        if m is None:
                            m = CFM[stRoot, size] = CombinationsFromTree(stRoot, size)
                        
                        S.append(m)
                        if not len(S[-1]):
                           fail = True
                           break
                    if fail:
                       continue
                    for combProduct in freduce(unionProduct, S):
                        lnodesets.add(frozenset(combProduct | {t}))
        return lnodesets

    CombinationTree(u, k, G_init)
    return {frozenset({treeL[f] for f in fs}) for fs in CombinationsFromTree(0, k)}

def compositions(n, k):
  if n < 0 or k < 0:
    return
  elif k == 0:
    if n == 0:
      yield []
    return
  elif k == 1:
    yield [n]
    return
  else:
    for i in range(1, n):
      for comp in compositions(n-i, k-1):
        yield [i] + comp



"""A python file that implements a class and functions for graphs."""

from __future__ import absolute_import
import numbers
import warnings

import numpy as np

from scipy.sparse import isspmatrix
from scipy.sparse.csgraph import laplacian

from .tools import inv_dict
from .tools import nested_dict_add
from .tools import priority_dict

# Python 2/3 cross-compatibility import
from six import iteritems
from six import itervalues
from six.moves.collections_abc import Iterable

from builtins import range


class Graph(object):
    """The general `graph` class.

    A general graph class that supports adjacency, dictionary formats
    while beeing memory/computationaly sustainable.

    Parameters
    ----------
    initialization_object :  dict, list or array-like, square, default=None
        The initialisation object for the graph (*valid-graph-format*).

            - If given a dictionary the input can be as follows:

                + 2-level nested dictionaries from edge symbols to weights

                + Dictionary of symbols to list of symbols (unweighted)

                + Dictionary of tuples to weights (weighted)

                + Iterable of tuples of len 2 (unweighted)

                + Iterable of tuples of len 3 (vertex, vertex, weight)

            - If given an array the input can be as follows:

                + array-like lists of lists

                + np.array

                + sparse matrix (scipy.sparse)

    node_labels : dict, default=None
        A label dictionary corresponding to all vertices of the graph:

            + for adjacency matrix labels should be given to numbers starting
              from 0 and ending in N-1, where the matrix has size N by N

            + for dictionary labels should correspond to all keys

    edge_labels : dict, default=None
        A labels dictionary corresponding to all edges of the graph
        with keys: index/symbol-tuples and value: labels.

    graph_format : str, valid_values={"dictionary", "adjacency", "all", "auto"}, default=None
        Defines the internal representation of the graph object which can be
        a dictionary as a matrix, or both:

            + for dictionary: "dictionary"

            + for adjacency_matrix: "adjacency"

            + for both: "all"

            + for the current_format (if existent): "auto"

    Attributes
    ----------
    n : int
        The one dimension of the (square) adjacency matrix.
        Signifies the number of vertices.

    adjacency_matrix : np.array
        The adjacency_matrix corresponding to the graph.

    index_node_labels : dict
        Label dictionary for indeces, of adjacency matrix.
        Keys are valid numbers from 0 to n-1.

    index_edge_labels : dict
        Label dictionary for edges, of adjacency matrix.
        Keys are tuple pairs of symbols with valid numbers from 0 to n-1,
        that have a positive adjacency matrix value.

    vertices : set
        The set of vertices corresponding to the edge_dictionary
        representation.

    edge_dictionary : dict
        A 2-level nested dictionary from edge symbols to weights.

    node_labels : dict
        Label dictionary for nodes.
        Keys are vertex symbols inside vertices.

    edge_labels : dict
        Label dictionary for edges.
        Keys are tuple pairs of symbols inside vertices and edges.

    edsamic : dict
        *Edge-Dictionary-Symbols-Adjacency-Matrix-Index-Correspondance*.
        A dictionary which translates dictionary symbols to adjacency
        matrix indexes, when storing both formats.

    shortest_path_mat : np.array, square
        Holds the shortest path matrix.

    label_group : dict
        A 2-level nested dict that after the first level of a pair tuple for
        purpose: "adjacency", "dictionary" and "vertex","edge" specification
        holds the inverse map of labels.

    laplacian_graph : np.array
        Holds the graph laplacian.

    _format : str, valid_values={"adjacency", "dictionary", "all"}
        Private attribute that keeps the current format.

    """

    # Adjacency Preset
    n = -1
    adjacency_matrix = np.empty(0)
    index_node_labels = dict()
    index_edge_labels = dict()

    # Dictionary Preset
    vertices = set()
    edge_dictionary = dict()
    node_labels = dict()
    edge_labels = dict()
    edsamic = dict()

    # Other Preset
    shortest_path_mat = None
    label_group = dict()
    laplacian_graph = None

    def __init__(
            self,
            initialization_object=None,
            node_labels=None, edge_labels=None,
            graph_format="auto", construct_labels=False):
        """__init__ function of the graph object."""
        self.construct_label = construct_labels
        if graph_format in ["adjacency", "dictionary", "auto", "all"]:
            self._format = graph_format
            if (initialization_object is not None):
                self.build_graph(initialization_object,
                                 node_labels,
                                 edge_labels)
            elif graph_format == "auto":
                raise ValueError('no initialization object ' +
                                 '- format must not be auto')
        else:
            raise ValueError('Invalid graph format.\nValid graph formats' +
                             ' are "all", "dictionary", "adjacency", "auto"')

    def build_graph(self, g, node_labels=None, edge_labels=None):
        """Build a graph structure, given a supported graph representation.

        Parameters
        ----------
        g : a valid graph format
            Similar to intialization object (of ``__init__``).

        node_labels: dict, default=None
            Node labels dictionary relevant to g format.

        edge_labels: dict, default=None
            Edge labels dictionary relevant to g format.

        Returns
        -------
        None.

        """
        self.shortest_path_mat = None
        self.label_group = None
        case = 0
        if g is not None:
            if is_adjacency(g):
                # Input is considered an adjacency matrix
                case = 1
                # Assign labels for nodes
                self.index_node_labels = node_labels

                # Assign labels for edges
                self.index_edge_labels = edge_labels

                if(self._format == "auto"):
                    self._format = "adjacency"
            elif is_edge_dictionary(g):
                # Input is considered as an edge dictionary
                case = 2

                # Assign labels for nodes
                self.node_labels = node_labels

                # Assign labels for edges
                self.edge_labels = edge_labels

                if(self._format == "auto"):
                    self._format = "dictionary"
            else:
                raise ValueError(
                    'Unsupported input type. For more information ' +
                    'check the documentation, concerning valid input ' +
                    'types for graph type object.')

        # If graph is of one type prune the other
        if self._format == "adjacency":
            self.edge_dictionary = None

        elif self._format == "dictionary":
            self.adjacency_matrix = None

        # Import the given input properly
        if (case == 1):
            self._import_adjacency(g)
        elif (case == 2):
            self._import_dictionary(g)

    def change_format(self, graph_format):
        """Change the format of the graph from an existing to an other.

        Parameters
        ----------
        graph_format : str, valid_values={"dictionary", "adjacency", "all"}
            Defines the internal representation of the graph object which can
            be a dictionary as a matrix, or both:

                + for dictionary: "dictionary"

                + for adjacency_matrix: "adjacency"

                + for both: "all"

        Returns
        -------
        None.

        """
        if graph_format not in ["all", "dictionary", "adjacency"]:
            raise ValueError(
                'Invalid graph format for function change format' +
                '. Valid formats are "all", "dictionary", "adjacency"')
        else:
            if (graph_format is not self._format):
                past_format = self._format
                self._format = graph_format
                if past_format == "adjacency":
                    self._import_adjacency()
                elif past_format == "dictionary":
                    self._import_dictionary()
                else:
                    if self._format == "dictionary":
                        self.n = -1
                        self.adjacency_matrix = None
                        self.edsamic = None
                        self.index_node_labels = None
                        self.index_edge_labels = None
                    else:
                        self.vertices = None
                        self.node_labels = None
                        self.edge_labels = None
                        self.edge_dictionary = None

    def desired_format(self, graph_format, warn=False):
        """Change the graph format, to include the desired.

        Parameters
        ----------
        graph_format : str, valid_values={"dictionary", "adjacency", "all"}
            Defines the internal representation of the graph object which
            can be a dictionary as a matrix, or both:

                + for dictionary: "dictionary"

                + for adjacency_matrix: "adjacency"

                + for both: "all"

        warn : bool, default=False
            Warn the user if the format of the graph is being changed.

        Returns
        -------
        None.

        """
        if graph_format == "all":
            self.change_format(graph_format)
        elif graph_format == "dictionary":
            if self._format not in ["all", "dictionary"]:
                if warn:
                    warnings.warn('changing format from "dictionary" to "all"')
                self.change_format("all")
        elif graph_format == "adjacency":
            if self._format not in ["all", "adjacency"]:
                warnings.warn('changing format from "adjacency" to "all"')
                self.change_format("all")

    def construct_labels(self, label_type="vertex", purpose="adjacency"):
        """Construct labels (if user does not provide).

        Parameters
        ----------
        label_type : str, valid_values={"vertex", "edge"}, default="vertex"
            What kind of labels are going to be constructed

        purpose : str, valid_values={"adjacency", "dictionary"}, default="adjacency"
            Defines if the labels correspond to dictionary or adjacency matrix

        Returns
        -------
        None.

        """
        if (purpose == "adjacency"):
            nodes = list(range(0, self.n))
            if (label_type == "vertex"):
                self.index_node_labels = dict(zip(nodes, nodes))
            elif (label_type == "edge"):
                self.index_edge_labels = {(i, j): self.adjacency_matrix[i, j]
                                          for i in nodes for j in nodes}
            else:
                warnings.warn('unsupported label type')
        elif (purpose == "dictionary"):
            nodes = sorted(list(self.vertices))
            if (label_type == "vertex"):
                self.node_labels = dict(zip(nodes, nodes))
            elif (label_type == "edge"):
                self.index_edge_labels = {
                    (i, j): (0 if (i not in self.edge_dictionary) or
                             (j not in self.edge_dictionary[i]) else
                             self.edge_dictionary[i][j])
                    for i in nodes for j in nodes}
            else:
                warnings.warn('unsupported label type')

    def convert_labels(
            self,
            target_format="dictionary",
            purpose="all",
            init=False):
        """Convert labels to a desired format.

        Parameters
        ----------
        target_format : str, valid_values={"adjacency", "dictionary"}, default="dictionary"
            Defines what the target format for conversion will be.

        purpose : str, valid_values={"adjacency", "dictionary", "all"}, default="all"
            Defines if the labels will be converted for dictionary, adjacency
            matrix or both.

        init : bool, default=False
            An override parameter for format checks,
            usefull for initialisation.

        Returns
        -------
        None.

        """
        if (target_format == "adjacency"):
            if (self._format != "adjacency" or init):
                if bool(self.index_node_labels) and \
                   purpose in ['all', 'vertex']:
                    warnings.warn('overriding existing' +
                                  'node labels for indexes')
                if bool(self.index_edge_labels) and \
                   purpose in ['all', 'edges']:
                    warnings.warn('overriding existing edge' +
                                  'labels for indexes')
                cond_labels_nodes = \
                    purpose in ['all', 'vertex'] and bool(self.node_labels)
                cond_labels_edges = \
                    purpose in ['all', 'edges'] and bool(self.edge_labels)
                if cond_labels_nodes or cond_labels_edges:
                    lov_sorted = sorted(list(self.vertices))
                    if cond_labels_nodes:
                        self.index_node_labels = \
                            {i: self.node_labels[k] for (i, k) in
                             enumerate(lov_sorted)}
                    if cond_labels_edges:
                        vi = {v: i for (i, v) in enumerate(lov_sorted)}
                        self.index_edge_labels = \
                            {(vi[k], vi[q]): self.edge_labels[(k, q)]
                             for (k, q) in self.edge_labels.keys()}
                else:
                    if not init:
                        warnings.warn('no labels to convert from, ' +
                                      'for the given purpose')
            else:
                warnings.warn('labels already defined for that format ' +
                              '- nothing to convert')
        elif (target_format == "dictionary"):
            if (self._format != "dictionary" or init):
                if bool(self.node_labels) and purpose in ['all', 'vertex']:
                    warnings.warn('overriding existing node labels, ' +
                                  'for dictionary symbols')

                if bool(self.edge_labels) and purpose in ['all', 'edges']:
                    warnings.warn('overriding existing edge labels, ' +
                                  'for dictionary symbols')

                cond_labels_nodes = \
                    purpose in ['all', 'vertex'] \
                    and bool(self.index_node_labels)
                cond_labels_edges = \
                    purpose in ['all', 'edges'] \
                    and bool(self.index_edge_labels)
                if cond_labels_nodes or cond_labels_edges:
                    if cond_labels_nodes:
                        self.node_labels = self.index_node_labels
                    if cond_labels_nodes:
                        self.edge_labels = self.index_edge_labels
                else:
                    if not init:
                        warnings.warn('no labels to convert from, ' +
                                      'for the given purpose')
            else:
                warnings.warn('labels already defined ' +
                              'for the given format')

    def label(self, obj, label_type="vertex", purpose="dictionary"):
        """Get the label of a vertex.

        Parameters
        ----------
        obj : hashable
            The candidate labeled object for the corresponding purpose
            and label type.

        label_type : str, valid_values={"vertex", "edge"}, default="vertex"
            Defines if the lookup of the label
            will be done for vertices or edges.

        purpose : str, valid_values={"dictionary", "adjacency", "any"}, default="dictionary"
            Defines if the lookup will be done on the existing
            ("any" - if "all" default is adjacency) to the "dictionary"
            or to the "adjacency" format of the graph.

        Returns
        -------
        label : str, list, ..., *valid-label-type*
            Returns the label of the current object on the defined lookup.

        """
        if purpose == "any":
            if self._format == "all":
                purpose = "adjacency"
            else:
                purpose = self._format

        if purpose not in ["dictionary", "adjacency"]:
            raise ValueError('unrecognized purpose')

        if (label_type == "vertex"):
            vertex = obj
            labels = self.get_labels(purpose=purpose, label_type=label_type)

            if labels is not None:
                if (vertex in labels):
                    return labels[vertex]
                else:
                    raise ValueError('no label assigned to this vertex symbol')
            else:
                warnings.warn('labels are not set - default value returned')
                return vertex
        elif (label_type == "edge"):
            edge = obj
            labels = self.get_labels(purpose=purpose, label_type=label_type)

            if not bool(labels):
                if (edge in labels):
                    return labels[edge]
                else:
                    raise ValueError('no label assigned to this edge symbol')
            else:
                warnings.warn('labels are not set - default value returned')
                if (edge[0] not in self.edge_dictionary) or \
                   (edge[1] not in self.edge_dictionary[edge[0]]):
                    return .0
                else:
                    return self.edge_dictionary[edge[0]][edge[1]]

    def relabel(self, new_labels, purpose="dictionary", label_type="vertex"):
        """Relabel the graph object, supporting the current format.

        Parameters
        ----------
        new_labels : dict
            The new labels corresponding to the label type and purpose.

        purpose : str, valid_values={"dictionary", "adjacency"}, default="dictionary"
            Defines if the new labels are given for
            "adjacency" or "dictionary".

        label_type : str, valid_values={"vertex", "edge"}, default="vertex"
            Defines if the new labels are for vertices or edges.

        Returns
        -------
        None.

        """
        # checks
        if purpose not in ["dictionary", "adjacency"]:
            raise ValueError('purpose can either be dictionary or adjacency')

        if label_type not in ["vertex", "edge"]:
            raise ValueError('label_type can either be vertex or edge')

        if new_labels is None or not bool(new_labels):
            warnings.warn('user must provide new labels, input is None')
        elif label_type == "vertex":
            if purpose == "dictionary":
                if self._format in ["dictionary", "all"]:
                    self.node_labels = new_labels
                else:
                    raise ValueError('graph is in a format that supports ' +
                                     'labels for dictionary symbols')

                # Transform to a valid format
                if self._format == "all":
                    self.convert_labels("adjacency", "vertex")
                if bool(self.label_group) and \
                        ((label_type, purpose) in self.label_group):
                    self.label_group[(label_type, purpose)] = dict()
            elif purpose == "adjacency":
                if self._format in ["all", "adjacency"]:
                    self.index_node_labels = new_labels
                else:
                    raise ValueError('graph is in a format that supports ' +
                                     'labels for indexes')

                if self._format == "all":
                    self.convert_labels("dictionary", "vertex")
                if bool(self.label_group) and \
                   (label_type, purpose) in self.label_group:
                    self.label_group[(label_type, purpose)] = dict()
            else:
                warnings.warn('no labels to convert from, ' +
                              'for the given purpose')
        elif label_type == "edge":
            if purpose == "dictionary":
                if self._format in ["dictionary", "all"]:
                    self.edge_labels = new_labels
                else:
                    raise ValueError('graph is in a format that ' +
                                     'supports labels for dictionary symbols')

                # Transform to a valid format
                if self._format == "all":
                    self.convert_labels("adjacency", "edge")
                if (label_type, purpose) in self.label_group:
                    self.label_group[(label_type, purpose)] = dict()

            elif purpose == "adjacency":
                if self._format in ["all", "adjacency"]:
                    self.index_edge_labels = new_labels
                else:
                    raise ValueError('graph is in a format that supports ' +
                                     'labels for indexes')

                if self._format == "all":
                    self.convert_labels("dictionary", "edge")
                if (label_type, purpose) in self.label_group:
                    self.label_group[(label_type, purpose)] = dict()

            else:
                warnings.warn('no labels to convert from, for ' +
                              'the given purpose')
        else:
            warnings.warn('unrecognized label type')

    def build_shortest_path_matrix(
            self,
            algorithm_type="auto",
            clean=False,
            labels="vertex"):
        """Build and return the shortest path matrix between all nodes.

        Parameters
        ----------
        algorithm_type : str, valid_values={"auto", "adjacency", "dictionary"}, default="auto"
            Defines which shortest-path algorithm will be used for building the
            shortest path matrix:

                + "dijkstra" : choses the dijkstra algorithm (Matrix
                  computation complexity: :math:`O(|V|(|E|+|V|)log(|V|))`

                + "floyd_warshall" : choses the floyd-warshall algorithm
                  (Matrix computation complexity: :math:`O(|V|^3)`

                + "auto" : choses the best possible algorithm for the current
                  format

        clean : bool, default=False
            Construct the shortest path matrix from scratch or output existing
            if exists

        labels : str, valid_values={"vertex", "edge", "all", "none"}, default="vertex"
            Returns labels corresponding for the indexes of the shortest path
            matrix for vertices, for edge (only for the valid ones on the
            original graph), for both ("all") and for no labels ("none")

        Returns
        -------
        shortest_path_matrix : np.array, shape=(:math:`|V|`, :math:`|V|`)
            The produced shortest path matrix.

        vertex_labels : dict
            The vertex labels, outputed only if *labels* parameter is either
            "vertex" or "all".

        edge_labels : dict
            The edge labels, outputed only if *labels* parameter is
            either "edge" or "all".

        """
        if labels not in ['vertex', 'edge', 'all', 'none']:
            raise ValueError('only labels for vertices and edges exist')

        if clean:
            self.shortest_path_mat = None

        if self.shortest_path_mat is not None:
            if labels == "all":
                return self.shortest_path_mat,\
                    self.get_labels(),\
                    self.get_labels("edge")
            if labels == "edge":
                return self.shortest_path_mat, self.get_labels("edge")
            if labels == "vertex":
                return self.shortest_path_mat, self.get_labels()
            else:
                return self.shortest_path_mat

        # Assign the desired algorithm
        if algorithm_type == "auto":
            if self._format in ["all", "dictionary"]:
                algorithm_type = "dijkstra"
            elif self._format == "adjacency":
                algorithm_type = "floyd_warshall"

        if algorithm_type == "dijkstra":
            # all format required
            self.desired_format("all", warn=True)
            if (bool(self.edsamic)):
                indexes = self.edsamic
            else:
                indexes = dict(zip(range(self.n), range(self.n)))

            # calculate shortest path matrix
            shortest_path_mat = np.full([self.n, self.n], float("Inf"))
            for k in indexes.keys():
                dict_fd, _ = dijkstra(self.edge_dictionary, k)
                for s in dict_fd.keys():
                    shortest_path_mat[indexes[k], indexes[s]] = dict_fd[s]

        elif algorithm_type == "floyd_warshall":
            self.desired_format("adjacency", warn=True)
            shortest_path_mat = floyd_warshall(self.adjacency_matrix)

        self.shortest_path_mat = shortest_path_mat
        if labels == "all":
            return shortest_path_mat,\
                self.get_labels(),\
                self.get_labels("edge")
        if labels == "edge":
            return shortest_path_mat, self.get_labels("edge")
        if labels == "vertex":
            return shortest_path_mat, self.get_labels()
        else:
            return shortest_path_mat

    def get_labels(self, label_type="vertex", purpose="adjacency", return_none=False):
        """Get labels corresponding to the purpose.

        Parameters
        ----------
        label_type : str, valid_values={"vertex", "edge"}, default="vertex"
            Defines if the the labels will correspond to vertices or edges.

        purpose : str, valid_values={"dictionary", "adjacency", "any"},
                  default="dictionary"
            Defines if the labels will correspond to "dictionary" (symbols),
            to "adjacency" indexes, or to "any" valid format (if "all" the
            result is for "adjacency").

        return_none : bool
            Defines if get_labels can return None in case of label absence, or raise an error.

        Returns
        -------
        labels : dict,
            Returns the labels for the given type and purpose.

        """
        if purpose == "adjacency":
            case = True
        elif purpose == "dictionary":
            case = False
        elif purpose == "any":
            if self._format in ['all', 'adjacency']:
                case = True
            else:
                case = False
        else:
            raise ValueError('unsupported label purpose')

        if case:
            self.desired_format("adjacency", warn=True)
            if label_type == "vertex":
                if not bool(self.index_node_labels):
                    if self.construct_label:
                        self.construct_labels(label_type, purpose)
                    else:
                        if return_none:
                            return None
                        else:
                            raise ValueError('Graph does not have any labels for vertices.')

                return self.index_node_labels
            elif label_type == "edge":
                if not bool(self.index_edge_labels):
                    if self.construct_label:
                        self.construct_labels(label_type, purpose)
                    else:
                        if return_none:
                            return None
                        else:
                            raise ValueError('Graph does not have any labels for edges.')
                return self.index_edge_labels
            else:
                raise ValueError('label type can only be "vertex" or "edge"')
        else:
            self.desired_format("dictionary", warn=True)
            if label_type == "vertex":
                if not bool(self.node_labels):
                    if self.construct_label:
                        self.construct_labels(label_type, purpose)
                    else:
                        if return_none:
                            return None
                        else:
                            raise ValueError('Graph does not have any labels for vertices.')
                return self.node_labels
            elif label_type == "edge":
                if not bool(self.edge_labels):
                    if self.construct_label:
                        self.construct_labels(label_type, purpose)
                    else:
                        if return_none:
                            return None
                        else:
                            raise ValueError('Graph does not have any labels for edges.')
                return self.edge_labels
            else:

                raise ValueError('label type can only be "vertex" or "edge"')

    def get_label_group(self, label_type="vertex", purpose="dictionary"):
        """Calculate the inverse dictionary for vertex labels (once).

        Parameters
        ----------
        label_type : str, valid_values={"vertex", "edge"}, default="vertex"
            Defines if the the labels-group will correspond
            to vertices or edges.

        purpose : str, valid_values={"dictionary", "adjacency", "any"},
                  default="dictionary"
            Defines if the labels-group will correspond to "dictionary"
            (symbols), to "adjacency" (indexes) or to "any" valid format
            (if "all" the result is for "adjacency").

        Returns
        -------
        label_group : dict
            Returns the inverse label group.

        """
        if not bool(self.label_group):
            self.label_group = dict()
        if not (label_type, purpose) in self.label_group:
            self.label_group[(label_type, purpose)] = dict()
        if not bool(self.label_group[(label_type, purpose)]):
            # calculate the label_group
            self.label_group[(label_type, purpose)] = \
                inv_dict(self.get_labels(label_type, purpose))
        return self.label_group[(label_type, purpose)]

    def neighbors(self, vertex, purpose="any", with_weights=False):
        """Find all neighbors of a vertex.

        Parameters
        ----------
        vertex : hashable
            The vertex, which neighbors we are searching for.

        purpose : str, valid_values={"adjacency", "dictionary", "any"},
                  default="any"
            Defines if the vertex is given for the "dictionary" format of
            the graph (symbol) to the "adjacency" (index) or to "any"
            existing format (if "all" the expected type is for "adjacency").

        with_weights : bool, default=False
            Defines if the neighbours will be outputed with weights.

        Returns
        -------
        neighbors : list or dict
            The neighbors of the given vertex.
                + with_weights=False: list of neighbor vertices

                + with_weights=True: dictionary between neighbor vertices and
                  edge labels

        """
        if purpose in ['adjacency', 'dictionary', "any"]:
            if purpose == 'dictionary':
                self.desired_format('dictionary')
                case = True
            if purpose == 'adjacency':
                self.desired_format('adjacency')
                case = False
            if purpose == "any":
                if self._format in ['all', 'adjacency']:
                    case = False
                else:
                    case = True
        else:
            raise ValueError('purpose is either "adjacency", ' +
                             '"dictionary" or "any"')

        if case:
            if vertex in self.edge_dictionary:
                if not with_weights:
                    return list(self.edge_dictionary[vertex].keys())
                else:
                    return self.edge_dictionary[vertex]
            else:
                raise ValueError('vertex not inside edge dictionary')
        else:
            idx = int(vertex)
            if 0 <= idx < self.n:
                out_idx = np.where(self.adjacency_matrix[idx, :] > 0)
                ns = out_idx[0].tolist()
                if not with_weights:
                    return ns
                else:
                    return dict(zip(ns,
                                    self.adjacency[idx, out_idx].tolist())[0])
            else:
                raise ValueError("item with index "+str(idx)+" does not exist")

    def _make_edsamic(self, vertices):
        """Produce the `edsamic`.

        `edsamic` is *edge-symbols-adjacency-matrix-correspondance-dictionary*.

        Parameters
        ----------
        vertices : list-able, sort-able
            The vertices of the graph.

        all_out : bool, default=True
            A variable that defines, if everything valuable will be outputed.

        Returns
        -------
        edsamic : dict
            The produced edsamic

        n : int
            The number of vertieces

        lov_sorted : list
            Sorted list of vertices.

        """
        if bool(vertices):
            # vertices to list
            lov = list(vertices)

            # length is adjacency matrix size
            n = len(lov)

            # sort lov indexes for labeling
            lov_sorted = sorted(lov)

            # edge_labels_adjacency_matrix_correspondance_dictionary
            edsamic = dict(zip(lov_sorted, range(n)))

            return edsamic, n, lov_sorted
        else:
            warnings.warn('wrong format: edge dictionary must exist')
            return None

    def _import_adjacency(self, adjacency_matrix=None, init=True):
        """Create a graph object representation, given its adjacency matrix.

        Parameters
        ----------
        adjacency_matrix : array-like, default=None
            If given a array the input can be as follows:
                + array-like lists of lists

                + np.array

                + sparse matrix (scipy.sparse.csr.csr_matrix)

            If None, imports the existing array, inside self.adjacency_matrix

        init : bool, default=True
            A parameter used to defined initialization.

        Returns
        -------
        None.

        """
        if adjacency_matrix is not None:
            # calculate graph size
            is_adj, adjacency_matrix = is_adjacency(adjacency_matrix, True)

            if not is_adj:
                raise ValueError('unsupported format type '
                                 'for adjacency matrix')

            n = adjacency_matrix.shape[0]

            if n != adjacency_matrix.shape[1]:
                raise ValueError('input matrix must be squared')

            # import_adjacency
            if (self._format == "all" or self._format == "adjacency"):
                self.adjacency_matrix = adjacency_matrix
            self.n = n
        else:
            n = self.n
            adjacency_matrix = self.adjacency_matrix

        # construct a dictionary out of the adjacency
        if self._format in ["all", "dictionary"]:
            vertices = set(range(n))

            self.vertices = vertices
            self.edge_dictionary = {i: dict() for i in range(n)}

            idx_i, idx_j = np.where(adjacency_matrix > 0)
            for (i, j) in zip(idx_i, idx_j):
                self.edge_dictionary[i][j] = adjacency_matrix[i, j]

            # Add labels
            self.convert_labels(target_format="dictionary",
                                purpose="all",
                                init=init)

            # Prune not interesting format
            if self._format == "dictionary":
                self.n = -1
                self.adjacency_matrix = None
                self.edsamic = None
                self.index_node_labels = None
                self.index_edge_labels = None
            else:
                self.edsamic = dict(zip(range(n), range(n)))

    def _import_dictionary(self, edge_dictionary=None, init=True):
        """Create a graph object representation given its edge dictionary.

        Parameters
        ----------
        edge_dictionary :  dict-like, default=None
            The edge_dictionary the input can be as follows:
                + 2-level nested dictionaries from edge symbols to weights

                + Dictionary of symbols to list of symbols (unweighted)

                + Dictionary of tuples to weights (weighted)

                + Iterable of tuples of len 2 (unweighted)

                + Iterable of tuples of len 3 (vertex, vertex, weight)

            If None, imports from the existing edge_dictionary

        init : bool, default=True
            A parameter used to defined initialization.

        Returns
        -------
        None.

        """
        if edge_dictionary is not None:
            # find vertices, refine dictionary
            vertices = set()
            edge_dictionary_refined = dict()
            is_edge_dict, vertices, edge_dictionary_refined =\
                is_edge_dictionary(edge_dictionary, True)
            if not is_edge_dict:
                raise ValueError('unsupported edge_dictionary format')

            # Save dictionary, vertices @self if needed
            if (self._format == "all" or self._format == "dictionary"):
                self.edge_dictionary = edge_dictionary_refined
            self.vertices = vertices
        else:
            vertices = self.vertices
            edge_dictionary_refined = self.edge_dictionary

        # Create and store the adjacency matrix
        if self._format in ["adjacency", "all"]:
            edsamic, self.n, lov_sorted = self._make_edsamic(vertices)
            # Initialize adjacency_matrix
            adjacency_matrix = np.zeros(shape=(self.n, self.n))

            # Produce and save adjacency matrix
            for k in edge_dictionary_refined.keys():
                for l in edge_dictionary_refined[k].keys():
                    adjacency_matrix[edsamic[k], edsamic[l]] = \
                        edge_dictionary_refined[k][l]

            self.adjacency_matrix = adjacency_matrix

            if self._format == "all":
                self.edsamic = edsamic

            # Add labels
            self.convert_labels(target_format="adjacency",
                                purpose="all",
                                init=init)

            # Prune for a certain format
            if self._format == "adjacency":
                self.edge_dictionary = None
                self.vertices = None
                self.node_labels = None
                self.edge_labels = None

    def laplacian(self, save=True):
        """Calculate the laplacian of the given graph.

        Parameters
        ----------
        save : bool, default=True
            Optional parameter to store the matrix.

        Returns
        -------
        laplacian : array-like
            Returns the graph laplacian

        """
        if self.laplacian_graph is not None:
            laplacian_graph = self.laplacian_graph
        else:
            self.desired_format("adjacency", warn=True)
            laplacian_graph = laplacian(self.adjacency_matrix)

            if save:
                self.laplacian_graph = laplacian_graph
        return laplacian_graph

    def get_vertices(self, purpose="adjacency"):
        """Create an iterable of vertices.

        Parameters
        ----------
        purpose : str, valid_values={"adjacency", "dictionary", "any"},
                  default="adjacency"
            Defines if the vertices will correspond given for the "dictionary"
            format of the graph (symbol) to the "adjacency" (index) or to "any"
            existing format (if "all" the expected type is for "adjacency").

        Returns
        -------
        vertices : iterable
            Returns an iterable on vertices

        """
        if purpose not in ["adjacency", "dictionary", "any"]:
            raise ValueError('purpose is either "adjacency" of "dictionary"')

        if purpose == "any":
            if self._format in ['all', 'adjacency']:
                purpose = "adjacency"
            else:
                purpose = "dictionary"

        if purpose == "adjacency":
            self.desired_format("adjacency", warn=True)
            return range(0, self.n)
        if purpose == "dictionary":
            self.desired_format("dictionary", warn=True)
            return self.vertices

    def get_edges(self, purpose="adjacency", with_weights=False):
        """Create an iterable of edges as tuples.

        Parameters
        ----------
        purpose : str, valid_values={"adjacency", "dictionary"}, default="adjacency"
            Defines if the edges is given for the "dictionary" format of the
            graph (symbol) to the "adjacency" (index).

        Returns
        -------
        vertices : list
            Returns a list of tuples for edges.

        """
        if purpose not in ["adjacency", "dictionary"]:
            raise ValueError('purpose is either "adjacency" of "dictionary"')

        if purpose == "adjacency":
            self.desired_format("adjacency", warn=True)
            idx_i, idx_j = np.where(self.adjacency_matrix > 0)
            edges = zip(idx_i, idx_j)
            if with_weights:
                return list(zip(edges, self.adjacency_matrix[idx_i, idx_j]))
            else:
                return list(edges)
        if purpose == "dictionary":
            self.desired_format("dictionary", warn=True)
            if with_weights:
                return [(i, j) for i in self.edge_dictionary.keys()
                        for j in self.edge_dictionary[i].keys()]
            else:
                return [((i, j), self.edge_dictionary[i][j])
                        for i in self.edge_dictionary.keys()
                        for j in self.edge_dictionary[i].keys()]

    def get_adjacency_matrix(self):
        """Get the adjacency matrix.

        Format agnostic method.

        Parameters
        ----------
        None.

        Returns
        -------
        adjacency_matrix : np.array
            Returns the adjacency matrix of the current graph.

        """
        if self._format == "dictionary":
            A = np.zeros(shape=(len(self.vertices), len(self.vertices)))
            v_map = {v: i for (i, v) in enumerate(sorted(list(self.vertices)))}
            for (k, v) in iteritems(self.edge_dictionary):
                i = v_map[k]
                for (kv, w) in iteritems(v):
                    A[i, v_map[kv]] = w
            return A
        else:
            return self.adjacency_matrix

    def get_edge_dictionary(self):
        """Get the edge dictionary.

        Format agnostic method.

        Parameters
        ----------
        None.

        Returns
        -------
        edge_dictionary : dict
            Returns the edge_dictionary of the current graph.

        """
        if self._format != "dictionary":
            idx_i, idx_j = np.where(self.adjacency_matrix > 0)
            edge_dictionary = {i: dict() for i in range(0, self.n)}
            for (i, j) in zip(idx_i, idx_j):
                edge_dictionary[i][j] = self.adjacency_matrix[i, j]
            return edge_dictionary
        else:
            return self.edge_dictionary

    def nv(self):
        """Get the number of vertices for any existing format.

        Parameters
        ----------
        None.

        Returns
        -------
        num_of_vertices : int
            Returns the number of vertices.

        """
        if self._format in ['all', 'adjacency']:
            return self.n
        else:
            return len(self.vertices)

    def produce_neighborhoods(
            self,
            r=3,
            purpose="adjacency",
            with_distances=False,
            d=-1,
            sort_neighbors=True):
        """Calculate neighborhoods for each node of a Graph, up to a depth.

        Parameters
        ----------
        r : int
            The neighborhood depth (radius).

        purpose : str, valid_values={"adjacency", "dictionary"},
                  default="adjacency"
            Defines if the neighborhood symbols will correspond given
            for the "dictionary" format of the graph (symbol) to the
            "adjacency" (index).

        with_distances : bool, default=False
            Defines if we need to calculate BFS distances for each pair.

        d : int, default=-1
            Maximum distance considered. If -1 is provided the distance is as
            max as the radius r.

        Returns
        -------
        N : dict
            A level, vertex nested dictionary of lists, corresponding to the
            neighbors of level :math:`l` for a certain vertex :math:`v`.

        D : dict
            For each level, set of tuples of nodes connected in that level.
            Appears only if with_distances is *True*.

        Dist_pair : dict
            A dictionary of all pairs and their distances.

        """
        N = dict()

        if purpose == "any":
            if self._format == "all":
                purpose = "adjacency"
            else:
                purpose = self._format

        if purpose not in ["adjacency", "dictionary"]:
            raise ValueError('purpose is either "adjacency" or "dictionary"')

        if r < 0:
            raise ValueError('r must be positive or equal to zero')

        if with_distances and d < 0:
            d = r
            warnings.warn('negative d as input - d set to r')

        # chain neighbors
        if sort_neighbors:
            def chain(n):
                return sorted(n)
        else:
            def chain(n):
                return n

        # Initialize neighborhoods
        vertices = list(self.get_vertices(purpose))
        N[0] = {i: {i} for i in vertices}

        # and distances
        if with_distances:
            D = dict()
            D[0] = set(zip(vertices, vertices))
            Dist_pair = {(v, v): 0 for v in vertices}

        if r > 0:
            N[1] = dict()
            if with_distances and d >= 1:
                D[1] = set()
            for i in vertices:
                ns = list(self.neighbors(i, purpose))
                N[1][i] = chain([i]+ns)
                if with_distances and d >= 1:
                    dset = {(i, n) for n in ns}
                    Dist_pair.update(zip(dset, len(dset)*[1]))
                    D[1] |= dset

            # calculate neighborhoods by a recursive formula
            # for all levels from 1 to r
            for level in range(1, max(r, d)):
                N[level+1] = dict()
                if with_distances and level <= d-1:
                    D[level+1] = set()
                for i in vertices:
                    neighbors = set()
                    for w in N[level][i]:
                        neighbors |= set(N[level][w])
                    N[level+1][i] = chain(list(neighbors))
                    if with_distances and level <= d-1:
                        dset = {(i, j)
                                for j in (neighbors - set(N[level][i]))}
                        Dist_pair.update(zip(dset, len(dset)*[level+1]))
                        D[level+1] |= dset
            if with_distances:
                for level in range(r+1, d):
                    N.pop(level, None)

        if with_distances:
            return N, D, Dist_pair
        else:
            return N

    def get_graph_object(self):
        """Return the graph object corresponding to 'any'.

        Format agnostic method.

        Parameters
        ----------
        None.

        Returns
        -------
        g : dict or np.array
            The graph Object.

        """
        if self._format in ["adjacency", "all"]:
            return self.adjacency_matrix
        else:
            return self.edge_dictionary

    def get_subgraph(self, vertices):
        """Calculate the subgraph of object in the same format as the original.

        Parameters
        ----------
        vertices : iterable
            An iterbale vertices extracted from the original graph.

        Returns
        -------
        subgraph : graph
            The induced subgraph, from the input vertices.

        """
        if vertices is None:
            raise ValueError('vertices must not be null')
        if type(vertices) in [str, int]:
            vertices = {vertices}
        else:
            vertices = set(vertices).copy()

        subgraph = Graph(graph_format=self._format)
        if self._format == 'adjacency':
            for v in vertices:
                if v < 0 or v >= self.n:
                    raise ValueError('vertices are not valid '
                                     'for the original graph format')
            recipe = [('enum_vertices', 'default'),
                      ('add_adjacency',)]

        elif self._format == 'dictionary':
            for v in vertices:
                if v not in self.edge_dictionary:
                    raise ValueError('vertices are not valid '
                                     'for the original graph format')

            recipe = [('add_edge_dict',)]
        else:
            fv = False
            fa = False
            for v in vertices:
                if not fv and v not in self.edge_dictionary:
                    fv = True
                if not fa and (v < 0 or v >= self.n):
                    fa = True
                if fa and fv:
                    raise ValueError('vertices are not valid for '
                                     'the original graph format')

            if not fa and fv:
                recipe = [('enum_vertices', 'default'),
                          ('get_correct', 'edsamic'),
                          ('add_adjacency',),
                          ('add_edge_dict',)]
            elif not fa:
                recipe = [('enum_vertices', 'default'),
                          ('get_correct', 'edsamic'),
                          ('add_adjacency',),
                          ('idxs_to_nodes',),
                          ('add_edge_dict',)]
            elif not fv:
                recipe = [('enum_vertices', 'edsamic'),
                          ('get_correct', 'edsamic'),
                          ('add_adjacency',),
                          ('add_edge_dict',)]

        def get_correct(i):
            return i

        for ingredient in recipe:
            if ingredient[0] == 'idxs_to_nodes':
                vertices = {self.edsamic[i] for i in vertices}

            elif ingredient[0] == 'enum_vertices':
                if ingredient[1] == 'edsamic':
                    inverse_edsamic = inv_dict(self.edsamic)
                    lov_sorted = sorted([int(inverse_edsamic[v][0])
                                         for v in vertices])
                    new_indexes, inv_new_indexes = {}, {}
                    for (i, l) in enumerate(lov_sorted):
                        new_indexes[i], inv_new_indexes[l] = l, i
                    subgraph.edsamic = {self.edsamic[inv_new_indexes[nik]]:
                                        nik for nik in new_indexes.keys()}
                else:
                    lov_sorted = sorted(list(vertices))
                    new_indexes = {l: i for (i, l) in enumerate(lov_sorted)}

            elif ingredient[0] == 'add_adjacency':
                subgraph.adjacency_matrix = self.adjacency_matrix[lov_sorted, :][:, lov_sorted]
                subgraph.n = len(new_indexes.keys())
                if bool(self.index_node_labels):
                    subgraph.index_node_labels = {
                        new_indexes[k]: self.index_node_labels[k]
                        for k in self.index_node_labels.keys()
                        if k in vertices}
                if bool(self.index_edge_labels):
                    subgraph.index_edge_labels = {
                        (new_indexes[i], new_indexes[j]):
                        self.index_edge_labels[(i, j)]
                        for (i, j) in self.index_edge_labels
                        if (i in vertices and j in vertices)}

            elif ingredient[0] == 'add_edge_dict':
                subgraph.edge_dictionary = {
                    get_correct(i): {
                        get_correct(j): self.edge_dictionary[i][j]
                        for j in self.edge_dictionary[i] if j in vertices}
                    for i in self.edge_dictionary if i in vertices}
                subgraph.vertices = vertices

                if bool(self.node_labels):
                    subgraph.node_labels = {
                        get_correct(k): self.node_labels[k]
                        for k in self.node_labels.keys() if k in vertices}
                if bool(self.edge_labels):
                    subgraph.index_edge_labels = {
                        (get_correct(i), get_correct(j)):
                            self.edge_labels[(i, j)]
                        for (i, j) in self.edge_labels
                        if (i in vertices and j in vertices)}
            elif ingredient[0] == 'get_correct':
                if ingredient[1] == 'edsamic':
                    def get_correct(i):
                        return self.edsamic[new_indexes[i]]

        return subgraph


def is_adjacency(g, transform=False):
    """Define if input is in a valid adjacency matrix format.

    Parameters
    ----------
    g : Object
        The input object.

    transform : bool, default=False
        Defines if the input will be transformed to the internal adjacency
        matrix support format.

    Returns
    -------
    is_adjacency : bool
        A variable that determines if the input is a valid adjacency matrix.

    g_transformed : np.array
        Holds the transformed object to an np.array.
        This output appears **only** if transform parameter is True.

    """
    if type(g) in [np.array, np.ndarray] and len(g.shape) == 2:
        if transform:
            return True, g
        else:
            return True
    elif isspmatrix(g):
        if transform:
            return True, g.todense()
        else:
            return True
    elif type(g) is list and all(
            isinstance(l, list) and
            all(isinstance(i, numbers.Number) for i in l) for l in g):
        if transform:
            return True, np.array(g)
        else:
            return True
    else:
        if transform:
            return False, None
        else:
            return False


def is_edge_dictionary(g, transform=False):
    """Define if input is in a valid edge dictionary format.

    Parameters
    ----------
    g : Object
        The input object.

    transform : bool, default=False
        Defines if the input will be transformed to the internal
        edge dictionary support format.

    Returns
    -------
    is_edge_dictionary : bool
        A variable that determines if the input is a valid edge dictionary.

    g_transformed_vertices : set
        Holds the transformed object vertices of the edge_dictionary.

    g_transformed_edge_dict : dict
        Holds the transformed object as a 2-level edge dict.
        This output appears **only** if transform parameter is True.

    """
    if type(g) is dict:
        if all(
                type(k) is tuple and len(k) == 2 and
                isinstance(n, numbers.Number) for (k, n) in iteritems(g)):
            if transform:
                vertices_key = set()
                vertices_val = set()
                edge_dict = dict()
                for (k, v) in iteritems(g):
                    vertices_key.add(k[0])
                    vertices_val.add(k[1])
                    nested_dict_add(edge_dict, v, k[0], k[1])

                # intialise empty edges
                v_dif = vertices_val - vertices_key
                if len(v_dif) > 0:
                    for v in v_dif:
                        edge_dict[v] = dict()
                return True, vertices_key | vertices_val, edge_dict
            else:
                return True
        if all(isinstance(d, list) for d in itervalues(g)):
            if transform:
                vertices_key = set()
                vertices_val = set()
                edge_dict = dict()
                for (k, v) in iteritems(g):
                    vertices_key.add(k)
                    vertices_val |= set(v)
                    for kp in v:
                        nested_dict_add(edge_dict, 1., k, kp)

                # intialise empty edges
                v_dif = vertices_val - vertices_key
                if len(v_dif) > 0:
                    for v in v_dif:
                        edge_dict[v] = dict()
                return True, vertices_key | vertices_val, edge_dict
            else:
                return True
        if all(
                isinstance(d, dict) and
                all(isinstance(n, numbers.Number)
                    for n in itervalues(d)) for d in itervalues(g)):
            if transform:
                vertices_key = set(g.keys())
                vertices_val = {kp for k in g.keys() for kp in g[k].keys()}
                v_dif = vertices_val - vertices_key

                # intialise empty edges
                if len(v_dif) > 0:
                    for v in v_dif:
                        g[v] = dict()
                return True, vertices_key | vertices_val, g
            else:
                return True
    if isinstance(g, Iterable):
        if all(type(t) is tuple and len(t) == 2 for t in g):
            if transform:
                vertices_key = set()
                vertices_val = set()
                edge_dict = dict()
                for (v, u) in g:
                    vertices_key.add(v)
                    vertices_val.add(u)
                    nested_dict_add(edge_dict, 1., v, u)
                v_dif = vertices_val - vertices_key

                # intialise empty edges
                if len(v_dif) > 0:
                    for v in v_dif:
                        edge_dict[v] = dict()
                return True, vertices_key | vertices_val, edge_dict
            else:
                return True
        elif all(type(t) is tuple and len(t) == 3 for t in g):
            if transform:
                vertices_key = set()
                vertices_val = set()
                edge_dict = dict()
                for (v, u, w) in g:
                    vertices_key.add(v)
                    vertices_val.add(u)
                    nested_dict_add(edge_dict, w, v, u)
                v_dif = vertices_val - vertices_key

                # intialise empty edges
                if len(v_dif) > 0:
                    for v in v_dif:
                        edge_dict[v] = dict()
                return True, vertices_key | vertices_val, edge_dict
            else:
                return True
    if transform:
        return False, None
    else:
        return False


def dijkstra(edge_dictionary, start_vertex, end_vertex=None):
    """Calculate the dijkstra algorithm `see`_.

    Parameters
    ----------
    edge_dictionary: dict
        A 2-level nested dictionary of symbols, with value corresponding to
        the weight.

    start_vertex: hashable
        The start vertex symbol
        (should exists as a key inside edge_dictionary).

    end_vertex: hashable
        The end vertex symbol (should exists as a key inside edge_dictionary).

    Returns
    -------
    dict_fd : dict
        The dictionary of final distances.

    dict_pred : dict
        The dictionary of predecessors.

    .. note::

        The majority of this function code came from :ref:`here`_.

        .. _here: http://code.activestate.com/recipes/
            119466-dijkstras-algorithm-for-shortest-paths/

    """
    dict_fd = {}    # dictionary of final distances
    dict_pred = {}    # dictionary of predecessors
    queue = priority_dict()   # est.dist. of non-final vert.
    queue[start_vertex] = 0

    for v in queue:
        dict_fd[v] = queue[v]
        if v == end_vertex:
            break

        for w in edge_dictionary[v]:
            vwLength = dict_fd[v] + edge_dictionary[v][w]
            if w in dict_fd:
                if vwLength < dict_fd[w]:
                    raise ValueError("Dijkstra: found better path to "
                                     "already-final vertex")
            elif w not in queue or vwLength < queue[w]:
                queue[w] = vwLength
                dict_pred[w] = v

    return dict_fd, dict_pred


def floyd_warshall(adjacency_matrix):
    """Calculate the Floyd Warshall, shortest path matrix.

    Parameters
    ----------
    adjacency_matrix : np.array, square
        The adjacency matrix of the graph, on which the distances are being
        calculated.

    Returns
    -------
    dist : np.array
        The shortest path matrix as produced by floyd warshall

    """
    n = adjacency_matrix.shape[0]

    # Initialization
    dist = np.array(adjacency_matrix, copy=True).astype(float)
    dist[dist == 0] = float("Inf")
    np.fill_diagonal(dist, 0)

    # Calculation
    for k in range(n):
        for i in range(n):
            dist[i, :] = np.minimum(dist[i, :], dist[i, k] + dist[k, :])

    return dist



from warnings import warn
from collections import Counter
from grakel import Kernel, Graph

# For python2/3 compatibility
from six.moves.collections_abc import Iterable


class VertexHistogram(Kernel):
    """Vertex Histogram kernel as found in :cite:`Sugiyama2015NIPS`

    Parameters
    ----------
    None.

    Attributes
    ----------
    None.

    """

    # Define the graph format that this kernel needs (if needed)
    # _graph_format = "auto" (default: "auto")

    def __init__(self,
                 n_jobs=n_jobs,
                 verbose=False,
                 normalize=False,
                 # kernel_param_1=kernel_param_1_default,
                 # ...
                 # kernel_param_n=kernel_param_n_default,
                 ):
        """Initialise an `odd_sth` kernel."""

        # Add new parameters
        self._valid_parameters |= new_parameters

        super(VertexHistogram, self).__init__(n_jobs=n_jobs, verbose=verbose, normalize=normalize)

        # Get parameters and check the new ones
        # @for i=1 to num_new_parameters
        #   self.kernel_param_i = kernel_param_i

        # self.initialized_.update({
        #    param_needing_initialization_1 : False
        #             ...
        #    param_needing_initialization_m : False
        # })

    def initialize_(self):
        """Initialize all transformer arguments, needing initialization."""
        # If you want to implement a parallelization by your self here is your chance
        # If there is a pairwise operation on the Kernel object there is parallelization is implemented
        # Just run the initialise from father to initialise a joblib Parallel (if n_jobs is not None).
        super(VertexHistogram, self).initialize_()

        # for i=1 .. m
        #     if not self.initialized_["param_needing_initialization_i"]:
        #         # Apply checks (raise ValueError or TypeError accordingly)
        #         # calculate derived fields stored on self._derived_field_ia .. z
        #         self.initialized_["param_needing_initialization_i"] = True
        pass

    def parse_input(self, X):
        """Parse and check the given input for vertex kernel.

        Parameters
        ----------
        X : iterable
            For the input to pass the test, we must have:
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given graph
            format).

        Returns
        -------
        out : list
            List of frequency-histogram for each Graph.

        """
        if not isinstance(X, Iterable):
            raise TypeError('input must be an iterable\n')
        else:
            out = list()
            for (i, x) in enumerate(iter(X)):
                is_iter = isinstance(x, Iterable)
                if is_iter:
                    x = list(x)
                if is_iter and len(x) in [0, 2, 3]:
                    if len(x) == 0:
                        warn('Ignoring empty element on index: '+str(i))
                        continue
                    else:
                        # Our element is an iterable of at least 2 elements
                        labels = x[1]
                elif type(x) is Graph:
                    # get labels in any existing format
                    labels = x.get_labels(purpose="any")
                else:
                    raise TypeError('each element of X must be either a ' +
                                     'graph object or a list with at least ' +
                                     'a graph like object and node labels ' +
                                     'dict \n')

                # Append frequencies for the current Graph
                out.append(Counter(labels.values()))

            if len(out) == 0:
                raise ValueError('parsed input is empty')
            return out

    def pairwise_operation(self, x, y):
        """Calculate sum of frequency products.

        Parameters
        ----------
        x, y : Counter
            Label-Frequency Counters as occur from `parse_input`.

        Returns
        -------
        kernel : number
            The kernel value.

        """
        return sum(x[k]*y[k] for k in x.keys())



"""The general `setup.py`  file."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
import sys
from platform import system

from setuptools import Extension, find_packages, setup
from numpy import get_include
from Cython.Build import build_ext

# Compile extensions

# Set optimization arguments for compilation
OS = system()
if OS == "Windows":
    extra_compile_args = ["/O2", "/w"]
elif OS in ["Linux", "Darwin"]:
    extra_compile_args = ["-O3", "-w"]

# Remove the "-Wstrict-prototypes" compiler option, which isn't valid for C++.
import distutils.sysconfig

cfg_vars = distutils.sysconfig.get_config_vars()
for key, value in cfg_vars.items():
    if type(value) == str:
        cfg_vars[key] = value.replace("-Wstrict-prototypes", "")

# Add the _c_functions extension on kernels
ext_address = "./grakel/kernels/_c_functions/"
ext = Extension(
    name="grakel.kernels._c_functions",
    sources=[
        ext_address + "functions.pyx",
        ext_address + "src/ArashPartov.cpp",
        ext_address + "src/sm_core.cpp",
    ],
    include_dirs=[ext_address + "include", get_include()],
    depends=[ext_address + "include/functions.hpp"],
    language="c++",
    extra_compile_args=extra_compile_args,
)

# Add the bliss library extension for calculating isomorphism
isodir = "./grakel/kernels/_isomorphism/"
blissdir = isodir + "bliss-0.50/"

# The essential bliss source files
blisssrcs = ["graph.cc", "heap.cc", "orbit.cc", "partition.cc", "uintseqhash.cc"]
blisssrcs = [blissdir + src for src in blisssrcs]
pn = str(sys.version_info[0])

# Compile intpybliss
intpybliss = Extension(
    name="grakel.kernels._isomorphism.intpybliss",
    define_macros=[("MAJOR_VERSION", "0"), ("MINOR_VERSION", "50beta")],
    include_dirs=[blissdir],
    language="c++",
    extra_compile_args=extra_compile_args,
    sources=[isodir + "intpyblissmodule_" + pn + ".cc"] + blisssrcs,
)

# Make bliss extension
bliss = Extension(
    name="grakel.kernels._isomorphism.bliss",
    include_dirs=[isodir],
    language="c++",
    extra_compile_args=extra_compile_args,
    sources=[isodir + "bliss.pyx"],
)

setup(
    packages=find_packages(),
    package_data={"grakel.tests": ["data/Cuneiform/*.txt", "data/MUTAG/*.txt"]},
    ext_modules=[intpybliss, bliss, ext],
    cmdclass={"build_ext": build_ext},
)



"""The Edge Histogram kernel as defined in :cite:`sugiyama2015halting`."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
from warnings import warn

from collections import Counter

from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from grakel.kernels import Kernel
from grakel.graph import Graph

from numpy import array
from numpy import zeros
from numpy import squeeze
from numpy import einsum
from scipy.sparse import csr_matrix

# Python 2/3 cross-compatibility import
from six import iteritems
from six import itervalues
from six.moves.collections_abc import Iterable


class EdgeHistogram(Kernel):
    """Edge Histogram kernel as found in :cite:`sugiyama2015halting`.

    Parameters
    ----------
    sparse : bool, or 'auto', default='auto'
        Defines if the data will be stored in a sparse format.
        Sparse format is slower, but less memory consuming and in some cases the only solution.
        If 'auto', uses a sparse matrix when the number of zeros is more than the half of the matrix size.
        In all cases if the dense matrix doesn't fit system memory, I sparse approach will be tried.

    Attributes
    ----------
    None.

    """

    def __init__(self, n_jobs=None, normalize=False, verbose=False, sparse='auto'):
        """Initialize an edge kernel."""
        super(EdgeHistogram, self).__init__(n_jobs=n_jobs, normalize=normalize, verbose=verbose)
        self.sparse = sparse
        self._initialized.update({'sparse': True})

    def initialize(self):
        """Initialize all transformer arguments, needing initialization."""
        if not self._initialized["n_jobs"]:
            if self.n_jobs is not None:
                warn('no implemented parallelization for EdgeHistogram')
            self._initialized["n_jobs"] = True
        if not self._initialized["sparse"]:
            if self.sparse not in ['auto', False, True]:
                TypeError('sparse could be False, True or auto')
            self._initialized["sparse"] = True

    def parse_input(self, X):
        """Parse and check the given input for EH kernel.

        Parameters
        ----------
        X : iterable
            For the input to pass the test, we must have:
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given graph
            format).

        Returns
        -------
        out : np.array, shape=(len(X), n_labels)
            A np array for frequency (cols) histograms for all Graphs (rows).

        """
        if not isinstance(X, Iterable):
            raise TypeError('input must be an iterable\n')
        else:
            rows, cols, data = list(), list(), list()
            if self._method_calling in [1, 2]:
                labels = dict()
                self._labels = labels
            elif self._method_calling == 3:
                labels = dict(self._labels)
            ni = 0
            for (i, x) in enumerate(iter(X)):
                is_iter = isinstance(x, Iterable)
                if is_iter:
                    x = list(x)
                if is_iter and len(x) in [0, 3]:
                    if len(x) == 0:
                        warn('Ignoring empty element on index: '+str(i))
                        continue
                    else:
                        # Our element is an iterable of at least 2 elements
                        L = x[2]
                elif type(x) is Graph:
                    # get labels in any existing format
                    L = x.get_labels(purpose="any", label_type="edge")
                else:
                    raise TypeError('each element of X must be either a ' +
                                    'graph object or a list with at least ' +
                                    'a graph like object and node labels ' +
                                    'dict \n')

                # construct the data input for the numpy array
                for (label, frequency) in iteritems(Counter(itervalues(L))):
                    # for the row that corresponds to that graph
                    rows.append(ni)

                    # and to the value that this label is indexed
                    col_idx = labels.get(label, None)
                    if col_idx is None:
                        # if not indexed, add the new index (the next)
                        col_idx = len(labels)
                        labels[label] = col_idx

                    # designate the certain column information
                    cols.append(col_idx)

                    # as well as the frequency value to data
                    data.append(frequency)
                ni += 1

            # Initialise the feature matrix
            if self._method_calling in [1, 2]:
                if self.sparse == 'auto':
                    self.sparse_ = (len(cols)/float(ni * len(labels)) <= 0.5)
                else:
                    self.sparse_ = bool(self.sparse)

            if self.sparse_:
                features = csr_matrix((data, (rows, cols)), shape=(ni, len(labels)), copy=False)
            else:
                # Initialise the feature matrix
                try:
                    features = zeros(shape=(ni, len(labels)))
                    features[rows, cols] = data
                except MemoryError:
                    warn('memory-error: switching to sparse')
                    self.sparse_, features = True, csr_matrix((data, (rows, cols)), shape=(ni, len(labels)), copy=False)

            if ni == 0:
                raise ValueError('parsed input is empty')
            return features

    def _calculate_kernel_matrix(self, Y=None):
        """Calculate the kernel matrix given a target_graph and a kernel.

        Each a matrix is calculated between all elements of Y on the rows and
        all elements of X on the columns.

        Parameters
        ----------
        Y : np.array, default=None
            The array between samples and features.

        Returns
        -------
        K : numpy array, shape = [n_targets, n_inputs]
            The kernel matrix: a calculation between all pairs of graphs
            between targets and inputs. If Y is None targets and inputs
            are the taken from self.X. Otherwise Y corresponds to targets
            and self.X to inputs.

        """
        if Y is None:
            K = self.X.dot(self.X.T)
        else:
            K = Y[:, :self.X.shape[1]].dot(self.X.T)

        if self.sparse_:
            return K.toarray()
        else:
            return K

    def diagonal(self):
        """Calculate the kernel matrix diagonal of the fitted data.

        Parameters
        ----------
        None.

        Returns
        -------
        X_diag : np.array
            The diagonal of the kernel matrix, of the fitted. This consists
            of each element calculated with itself.


        """
        # Check is fit had been called
        check_is_fitted(self, ['X', 'sparse_'])
        try:
            check_is_fitted(self, ['_X_diag'])
        except NotFittedError:
            # Calculate diagonal of X
            if self.sparse_:
                self._X_diag = squeeze(array(self.X.multiply(self.X).sum(axis=1)))
            else:
                self._X_diag = einsum('ij,ij->i', self.X, self.X)
        try:
            # If transform has happened return both diagonals
            check_is_fitted(self, ['_Y'])
            if self.sparse_:
                Y_diag = squeeze(array(self._Y.multiply(self._Y).sum(axis=1)))
            else:
                Y_diag = einsum('ij,ij->i', self._Y, self._Y)
            return self._X_diag, Y_diag
        except NotFittedError:
            # Else just return both X_diag
            return self._X_diag



"""The main graph kernel class, implemented as a sci-kit transformer."""
import copy
import warnings

import numpy as np

from scipy.linalg import svd
from sklearn.base import BaseEstimator
from sklearn.base import TransformerMixin
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted

from grakel.kernels import GraphletSampling
from grakel.kernels import RandomWalk
from grakel.kernels import RandomWalkLabeled
from grakel.kernels import ShortestPath
from grakel.kernels import ShortestPathAttr
from grakel.kernels import WeisfeilerLehman
from grakel.kernels import NeighborhoodHash
from grakel.kernels import PyramidMatch
from grakel.kernels import SubgraphMatching
from grakel.kernels import NeighborhoodSubgraphPairwiseDistance
from grakel.kernels import LovaszTheta
from grakel.kernels import SvmTheta
from grakel.kernels import OddSth
from grakel.kernels import Propagation
from grakel.kernels import PropagationAttr
from grakel.kernels import HadamardCode
from grakel.kernels import MultiscaleLaplacian
from grakel.kernels import VertexHistogram
from grakel.kernels import EdgeHistogram
from grakel.kernels import GraphHopper
from grakel.kernels import CoreFramework
from grakel.kernels import WeisfeilerLehmanOptimalAssignment

# Python 2/3 cross-compatibility import
from future.utils import iteritems

# Supported base kernels
sbk = [
    ["vertex_histogram", "subtree_wl", "VH", "ST-WL"],
    ["edge_histogram", "EH"],
    ["random_walk", "RW"],
    ["shortest_path", "SP"],
    ["graphlet_sampling", "GR"],
    ["subgraph_matching", "SM"],
    ["multiscale_laplacian", "ML"],
    ["lovasz_theta", "LOVT"],
    ["svm_theta", "SVMT"],
    ["neighborhood_hash", "NH"],
    ["neighborhood_subgraph_pairwise_distance", "NSPD"],
    ["odd_sth", "ODD"],
    ["propagation", "PR"],
    ["pyramid_match", "PM"],
    ["graph_hopper", "GH"],
    ["weisfeiler_lehman_optimal_assignment", "WL-OA"]
    ]

sbks = set(e for ls in sbk for e in ls)

# Supported frameworks
sf = [
    ["weisfeiler_lehman", "WL"],
    ["hadamard_code", "HC"],
    ["core_framework", "CORE"]
    ]

sfs = set(e for ls in sf for e in ls)

# Supported kernels message
sep = u"\n \u27E1 "
sk_msg = ("Base-Kernels\n" + 12*"-" + sep + sep.join(','.join(synonyms) for synonyms in sbk) +
          "\n\nFrameworks\n" + 10*"-" + sep + sep.join(','.join(synonyms) for synonyms in sf))

# Defaults
default_n_components = 100


class GraphKernel(BaseEstimator, TransformerMixin):
    r"""A generic wrapper for graph kernels.

    Parameters
    ----------
    kernel : list*(dict or str)
        A single element or a list of :code:`dict` with:

            * "name" : [str] - with the kernel name

            * "name_of_parameter_1" : value

            * "name_of_parameter_2" : value

            * :math:`\;\cdots\;`

            * "name_of_parameter_k" : value

        or of :code:`str`, designating a kernel name.

        available "name" or "name-alias" / "parametres" are:
            1. base_graph_kernels (the structure must always reach a base kernel)

                - "vertex_histogram" or "subtree_wl" or "VH" or "ST-WL"
                    + (**o**) "sparse" : bool or 'auto'

                - "edge_histogram" or "EH"
                    + (**o**) "sparse" : bool or 'auto'

                - "random_walk" or "RW"
                    + (**o**) "with_labels" : bool

                    + (**o**) "lamda" : float

                    + (**o**) "method_type" : [str], "baseline", "fast"

                    + (**o**) "kernel_type" : [str], "geometric", "exponential"

                    + (**o**) "p" : [int] > 0


                - "shortest_path" or "SP"
                    + (**o**) "algorithm_type" : [str] "dijkstra", "floyd_warshall"

                    + (**o**) "as_attributes" : [bool]

                    + (**o**) "metric" : [function] : (attribute_x, attribute_y) -> number

                    + (**o**) "with_labels" : [bool]

                - "graphlet_sampling" or "GR"
                    + (**o**) "k" : [int]

                    + (**o**) "sampling" : [dict] or **None**

                - "multiscale_laplacian" or "ML"
                    + (**o**) "L" : [int] > 0

                    + (**o**) "gamma" : [float] > .0

                    + (**o**) "heta" : [float] > .0

                    + (**o**) "n_samples" : [int] > 0, if "which": "fast"

                    + (**o**) "P" : [int] > 0, if "which": "fast"

                - "subgraph_matching" or "SM"
                    + (**o**) "kv" : [function] : (node_x, node_y, Lx, Ly) -> number

                    + (**o**) "ke" : [function] : (edge_x, edge_y, Lx, Ly) -> number

                    + (**o**) "lw" : a lambda weight function for cliques: set -> number

                - "lovasz_theta" or "LOVT"
                    + (**o**) "n_samples" : [int] > 1

                    + (**o**) "subsets_size_range" : [tuple] of two [int]

                    + (**o**) "metric" : [function] (number, number) -> number

                - "svm_theta" or "SVMT"
                    + (**o**) "n_samples" : [int] > 1

                    + (**o**) "subsets_size_range" : [tuple] with 2 [int] elements

                    + (**o**) "metric" : [function] (number, number) -> number

                - "neighborhood_hash" or "NH"
                    + (**o**) "nh_type" : [str] "simple" or "count-sensitive"

                    + (**o**) "R" : [int] > 0

                    + (**o**) "bits" : [int] > 0

                - "neighborhood_subgraph_pairwise_distance" or "NSPD"
                    + (**o**) "r" : (int) positive integer

                    + (**o**) "d" : (int) positive integer

                - "odd_sth" or "ODD"
                    + (**o**) "h" : [int] > 0

                - "propagation" or "PR"
                    + (**o**) t_max: [int] > 0

                    + (**o**) T: [dict] [int]: [np.arrays]

                    + (**o**) with_attributes: [bool], default=False

                    + (**o**) M: [str] {"H", "TV"} if `with_attributes=True` else {"L1", "L2"}

                    + (**o**) w: [int] > 0

                    + (**o**) metric: [function] x:[Counter] , y:[Counter] -> [number]

                - "pyramid_match" or "PM"
                    + (**o**) with_labels: [bool]

                    + (**o**) d: [int] > 0

                    + (**o**) L: [int] >= 0

                - "graph_hopper" or "GH"
                    + (**o**) kernel_type: [str: {'linear', 'gaussian'}] or [tuple: {('gaussian', mu)}]
                      or [function] x:[(np.array, np.array)] , y:[(np.array, np.array)] -> [number]

                - "weisfeiler_lehman_optimal_assignment" or "WL-OA"
                    + (**o**) "n_iter" : [int] >= 0

            2. frameworks (if a next kernel in the list it asssigned as a base-kernel, else see default)
                - "weisfeiler_lehman" or "WL" / default="VH"
                    + (**o**) "n_iter" : [int] >= 0

                - "hadamard_code" or "HC" / default="VH"
                    + (**o**) "n_iter" : [int] > 0

                - "core_framework" or "CORE" / default="SP"
                    + (**o**) "min_core" : [int] >= -1

        where (**o**): stands for optional parameters

    Nystroem : int or bool, optional
        Defines the number of nystroem components.
        To initialize the default (100 components), set -1 or 0.

    n_jobs : int or None, optional
        Defines the number of jobs of a joblib.Parallel objects needed for parallelization
        or None for direct execution. The use or not of this function depends on each kernel.

    normalize : bool, optional
        Normalize the output of the graph kernel.
        Ignored when Nystroem GraphKernel object is instanciated.

    verbose : bool, optional
        Define if messages will be printed on stdout.

    random_state :  RandomState or int, default=None
        A random number generator instance or an int to initialize a RandomState as a seed.

    Attributes
    ----------
    _initialized : dict
        Monitors which parameter derived object should be _initialized.

    kernel_ : function
        The full kernel applied between graph objects.

    nystroem_ : int
        Holds the nystroem, number of components.
        If not _initialized, it stands as a False
        boolean variable.

    components_ : array, shape=(n_components, n_features)
        Subset of training graphs used to construct the feature map.

    nystroem_normalization_ : array, shape=(n_components, n_components)
        Normalization matrix needed for embedding.
        Square root of the kernel matrix on ``components_``.

    component_indices_ : array, shape=(n_components)
        Indices of ``components_`` in the training set.

    random_state_ : RandomState
        A RandomState object handling all randomness of the class.

    """

    def __init__(self,
                 kernel="shortest_path",
                 normalize=False,
                 verbose=False,
                 n_jobs=None,
                 random_state=None,
                 Nystroem=False):
        """`__init__` for `GraphKernel` object."""
        self.kernel = kernel
        self.normalize = normalize
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.Nystroem = Nystroem
        self._initialized = {"kernel": False,
                             "Nystroem": False,
                             "random_state": False,
                             "normalize": False,
                             "verbose": False,
                             "n_jobs": False}

    def fit(self, X, y=None):
        """Fit a dataset, for a transformer.

        Parameters
        ----------
        X : iterable
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given grap
            format). The train samples.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        self : object
            Returns self.

        """
        # Initialize the Graph Kernel.
        self.initialize()

        # Input validation and parsing
        if bool(self.nystroem_):
            X = list(X)
            nx = len(X)
            # get basis vectors
            if self.nystroem_ > nx:
                n_components = nx
                warnings.warn("n_components > n_samples. This is not "
                              "possible.\nn_components was set to n_samples"
                              ", which results in inefficient evaluation of"
                              " the full kernel.")
            else:
                n_components = self.nystroem_

            n_components = min(nx, n_components)
            inds = self.random_state_.permutation(nx)
            basis_inds = inds[:n_components]
            basis = [X[i] for i in basis_inds]

            # sqrt of kernel matrix on basis vectors
            U, S, V = svd(self.kernel_.fit_transform(basis))
            S = np.maximum(S, 1e-12)
            self.nystroem_ = n_components
            self.nystroem_normalization_ = np.dot(U / np.sqrt(S), V)
            self.components_ = basis
            self.component_indices_ = inds
        else:
            self.kernel_.fit(X)

        # Return the transformer
        return self

    def transform(self, X):
        """Calculate the kernel matrix, between given and fitted dataset.

        Parameters
        ----------
        X : iterable
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given graph
            format). If None the kernel matrix is calculated upon fit data.
            The test samples.

        Returns
        -------
        K : numpy array, shape = [n_targets, n_input_graphs]
            corresponding to the kernel matrix, a calculation between
            all pairs of graphs between target an features

        """
        # Transform - calculate kernel matrix
        check_is_fitted(self, 'kernel_')
        if hasattr(self, 'nystroem_') and bool(self.nystroem_):
            # Check if nystroem has been initialized had been called
            check_is_fitted(self, 'components_')
            K = self.kernel_.transform(X).dot(self.nystroem_normalization_.T)
        else:
            K = self.kernel_.transform(X)

        return K

    def fit_transform(self, X, y=None):
        """Fit and transform, on the same dataset.

        Parameters
        ----------
        X : iterable
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given graph
            format). If None the kernel matrix is calculated upon fit data.
            The test samples.

        y : None
            There is no need of a target in a transformer, yet the pipeline API
            requires this parameter.

        Returns
        -------
        K : numpy array, shape = [n_targets, n_input_graphs]
            corresponding to the kernel matrix, a calculation between
            all pairs of graphs between target an features

        """
        # Initialize the Graph Kernel
        self.initialize()

        # Transform - calculate kernel matrix
        if bool(self.nystroem_):
            self.fit(X)
            K = self.kernel_.transform(X).dot(self.nystroem_normalization_.T)
        else:
            K = self.kernel_.fit_transform(X)

        return K

    def initialize(self):
        """Initialize all transformer arguments, needing initialisation."""
        if not self._initialized["Nystroem"]:
            if type(self.Nystroem) not in [int, bool]:
                raise ValueError('Nystroem parameter must be an int, '
                                 'indicating the number of components'
                                 'or a boolean')
            elif self.Nystroem is False:
                self.nystroem_ = False
            elif self.Nystroem in [0, -1] or self.Nystroem is True:
                # picking default number of components
                self.nystroem_ = default_n_components
            elif self.Nystroem <= 0:
                raise ValueError('number of nystroem components '
                                 'must be positive')
            else:
                self.nystroem_ = self.Nystroem
            self._initialized["Nystroem"] = True

        if any(not self._initialized[param]
               for param in ["random_state", "normalize", "verbose", "n_jobs", "kernel"]):
            # Intialise random_state_
            if not self._initialized["random_state"]:
                self.random_state_ = check_random_state(self.random_state)

            k = self.kernel
            if type(k) is dict or type(k) is str:
                # allow single kernel dictionary inputs
                k = [self.kernel]
            elif type(k) is not list:
                raise ValueError('A "kernel" must be defined at the __init__ '
                                 'function of the graph kernel generic wrapper.'
                                 'Valid kernel types are dict, str, and list of dict or str.')

            hidden_args = {"verbose": self.verbose, "normalize": self.normalize,
                           "n_jobs": self.n_jobs}

            # Initialize a new kernel each time a new fit is being called
            kernel, params = self.make_kernel_(copy.deepcopy(k), hidden_args)
            self.kernel_ = kernel(**params)
            for param in ["random_state", "normalize", "verbose", "n_jobs", "kernel"]:
                self._initialized[param] = True

    def make_kernel_(self, kernel_list, hidden_args):
        """Produce the desired kernel function.

        Parameters
        ----------
        kernel_list : (list)
            List of kernel dictionaries as defined at the documentation
            of class parameters.

        Returns
        -------
        kernel : kernel (class).
            Returns an instance of a kernel type object corresponding to the
            certain kernel.

        """
        kernel = kernel_list.pop(0)
        if type(kernel) is str:
            kernel_name, kernel = str(kernel), dict()
        elif type(kernel) is not dict:
            raise ValueError('each element of the list of kernels must be a dictionary or a string')
        else:
            if "name" not in kernel:
                raise ValueError('each dictionary concerning a kernel must '
                                 'have a "name" parameter designating the '
                                 'kernel')
            kernel_name = kernel.pop("name")

        for (keys, val) in iteritems(hidden_args):
            if keys in kernel:
                warnings.warn('Overriding global kernel attribute ' + str(keys) + ' with ' + str(val) +
                              '. Please set this attribute as an argument of GraphKernel.')
            kernel[keys] = val

        def get_random_state_(kernel):
            return kernel.pop(
                "random_state",
                (self.random_state_ if self.random_state is not None else None))

        if kernel_name in sbks:
            if len(kernel_list) != 0:
                warnings.warn('Kernel List not empty while reaching a base-kernel - '
                              'the rest kernel names will be ignored')

            if kernel_name in sbk[0]:
                return VertexHistogram, kernel
            elif kernel_name in sbk[1]:
                return EdgeHistogram, kernel
            elif kernel_name in sbk[2]:
                if kernel.pop("with_labels", False):
                    return RandomWalkLabeled, kernel
                else:
                    return RandomWalk, kernel
            elif kernel_name in sbk[3]:
                if kernel.pop("as_attributes", False):
                    return ShortestPathAttr, kernel
                else:
                    return ShortestPath, kernel
            elif kernel_name in sbk[4]:
                kernel["random_state"] = get_random_state_(kernel)
                return GraphletSampling, kernel
            elif kernel_name in sbk[5]:
                return SubgraphMatching, kernel
            elif kernel_name in sbk[6]:
                kernel["random_state"] = get_random_state_(kernel)
                return (MultiscaleLaplacian, kernel)
            elif kernel_name in sbk[7]:
                kernel["random_state"] = get_random_state_(kernel)
                return LovaszTheta, kernel
            elif kernel_name in sbk[8]:
                kernel["random_state"] = get_random_state_(kernel)
                return SvmTheta, kernel
            elif kernel_name in sbk[9]:
                return NeighborhoodHash, kernel
            elif kernel_name in sbk[10]:
                return NeighborhoodSubgraphPairwiseDistance, kernel
            elif kernel_name in sbk[11]:
                return OddSth, kernel
            elif kernel_name in sbk[12]:
                kernel["random_state"] = get_random_state_(kernel)
                if kernel.pop("with_attributes", False):
                    return PropagationAttr, kernel
                else:
                    return Propagation, kernel
            elif kernel_name in sbk[13]:
                return PyramidMatch, kernel
            elif kernel_name in sbk[14]:
                return GraphHopper, kernel
            elif kernel_name in sbk[15]:
                return WeisfeilerLehmanOptimalAssignment, kernel

        elif kernel_name in sfs:
            if len(kernel_list):
                kernel["base_graph_kernel"] = self.make_kernel_(kernel_list, {})
            if kernel_name in sf[0]:
                return (WeisfeilerLehman, kernel)
            elif kernel_name in sf[1]:
                return (HadamardCode, kernel)
            elif kernel_name in sf[2]:
                return (CoreFramework, kernel)
        else:
            raise ValueError("Unsupported kernel: " + str(kernel_name) + "\n"
                             "Supported kernels are:\n\n" + sk_msg)

    def set_params(self, **params):
        """Call the parent method."""
        # Copy the parameters
        params = copy.deepcopy(params)

        # Iterate over the parameters
        for key, value in iteritems(params):
            key, delim, sub_key = key.partition('__')
            if delim:
                if sub_key in self._initialized:
                    self._initialized[sub_key] = False
            elif key in self._initialized:
                self._initialized[key] = False

        # Set parameters
        super(GraphKernel, self).set_params(**params)



"""Neighborhood subgraph pairwise distance kernel :cite:`costa2010fast`."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
import warnings
from collections import defaultdict

import numpy as np

from scipy.sparse import csr_matrix

from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from grakel.kernels import Kernel
from grakel.graph import Graph

from grakel.kernels._c_functions import APHash

# Python 2/3 cross-compatibility import
from six import iteritems
from six.moves import filterfalse
from builtins import range
from six.moves.collections_abc import Iterable


class NeighborhoodSubgraphPairwiseDistance(Kernel):
    """The Neighborhood subgraph pairwise distance kernel.

    See :cite:`costa2010fast`.

    Parameters
    ----------
    r : int, default=3
        The maximum considered radius between vertices.

    d : int, default=4
        Neighborhood depth.

    Attributes
    ----------
    _ngx : int
        The number of graphs upon fit.

    _ngy : int
        The number of graphs upon transform.

    _fit_keys : dict
        A dictionary with keys from `0` to `_d+1`, constructed upon fit
        holding an enumeration of all the found (in the fit dataset)
        tuples of two hashes and a radius in this certain level.

    _X_level_norm_factor : dict
        A dictionary with keys from `0` to `_d+1`, that holds the self
        calculated kernel `[krg(X_i, X_i) for i=1:ngraphs_X]` for all levels.

    """

    _graph_format = "dictionary"

    def __init__(self, n_jobs=None, normalize=False, verbose=False, r=3, d=4):
        """Initialize an NSPD kernel."""
        # setup valid parameters and initialise from parent
        super(NeighborhoodSubgraphPairwiseDistance, self).__init__(
            n_jobs=n_jobs,
            normalize=normalize,
            verbose=verbose)

        self.r = r
        self.d = d
        self._initialized.update({"r": False, "d": False})

    def initialize(self):
        """Initialize all transformer arguments, needing initialization."""
        if not self._initialized["n_jobs"]:
            if self.n_jobs is not None:
                warnings.warn('no implemented parallelization for NeighborhoodSubgraphPairwiseDistance')
            self._initialized["n_jobs"] = True

        if not self._initialized["r"]:
            if type(self.r) is not int or self.r < 0:
                raise ValueError('r must be a positive integer')
            self._initialized["r"] = True

        if not self._initialized["d"]:
            if type(self.d) is not int or self.d < 0:
                raise ValueError('d must be a positive integer')
            self._initialized["d"] = True

    def parse_input(self, X):
        """Parse and create features for the NSPD kernel.

        Parameters
        ----------
        X : iterable
            For the input to pass the test, we must have:
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that correspond to the given
            graph format). A valid input also consists of graph type objects.

        Returns
        -------
        M : dict
            A dictionary with keys all the distances from 0 to self.d
            and values the the np.arrays with rows corresponding to the
            non-null input graphs and columns to the enumerations of tuples
            consisting of pairs of hash values and radius, from all the given
            graphs of the input (plus the fitted one's on transform).

        """
        if not isinstance(X, Iterable):
            raise TypeError('input must be an iterable\n')
        else:
            # Hold the number of graphs
            ng = 0

            # Holds all the data for combinations of r, d
            data = defaultdict(dict)

            # Index all keys for combinations of r, d
            all_keys = defaultdict(dict)
            for (idx, x) in enumerate(iter(X)):
                is_iter = False
                if isinstance(x, Iterable):
                    is_iter, x = True, list(x)
                if is_iter and len(x) in [0, 3]:
                    if len(x) == 0:
                        warnings.warn('Ignoring empty element' +
                                      ' on index: '+str(idx))
                        continue
                    else:
                        g = Graph(x[0], x[1], x[2])
                        g.change_format("adjacency")
                elif type(x) is Graph:
                    g = Graph(x.get_adjacency_matrix(),
                              x.get_labels(purpose="adjacency", label_type="vertex"),
                              x.get_labels(purpose="adjacency", label_type="edge"))
                else:
                    raise TypeError('each element of X must have either ' +
                                    'a graph with labels for node and edge ' +
                                    'or 3 elements consisting of a graph ' +
                                    'type object, labels for vertices and ' +
                                    'labels for edges.')

                # Bring to the desired format
                g.change_format(self._graph_format)

                # Take the vertices
                vertices = set(g.get_vertices(purpose=self._graph_format))

                # Extract the dicitionary
                ed = g.get_edge_dictionary()

                # Convert edges to tuples
                edges = {(j, k) for j in ed.keys() for k in ed[j].keys()}

                # Extract labels for nodes
                Lv = g.get_labels(purpose=self._graph_format)
                # and for edges
                Le = g.get_labels(purpose=self._graph_format, label_type="edge")

                # Produce all the neighborhoods and the distance pairs
                # up to the desired radius and maximum distance
                N, D, D_pair = g.produce_neighborhoods(self.r, purpose="dictionary",
                                                       with_distances=True, d=self.d)

                # Hash all the neighborhoods
                H = self._hash_neighborhoods(vertices, edges, Lv, Le, N, D_pair)

                if self._method_calling == 1:
                    for d in filterfalse(lambda x: x not in D,
                                         range(self.d+1)):
                        for (A, B) in D[d]:
                            for r in range(self.r+1):
                                key = (H[r, A], H[r, B])
                                keys = all_keys[r, d]
                                idx = keys.get(key, None)
                                if idx is None:
                                    idx = len(keys)
                                    keys[key] = idx
                                data[r, d][ng, idx] = data[r, d].get((ng, idx), 0) + 1

                elif self._method_calling == 3:
                    for d in filterfalse(lambda x: x not in D,
                                         range(self.d+1)):
                        for (A, B) in D[d]:
                            # Based on the edges of the bidirected graph
                            for r in range(self.r+1):
                                keys = all_keys[r, d]
                                fit_keys = self._fit_keys[r, d]
                                key = (H[r, A], H[r, B])
                                idx = fit_keys.get(key, None)
                                if idx is None:
                                    idx = keys.get(key, None)
                                    if idx is None:
                                        idx = len(keys) + len(fit_keys)
                                        keys[key] = idx
                                data[r, d][ng, idx] = data[r, d].get((ng, idx), 0) + 1
                ng += 1
            if ng == 0:
                raise ValueError('parsed input is empty')

            if self._method_calling == 1:
                # A feature matrix for all levels
                M = dict()

                for (key, d) in filterfalse(lambda a: len(a[1]) == 0,
                                            iteritems(data)):
                    indexes, data = zip(*iteritems(d))
                    rows, cols = zip(*indexes)
                    M[key] = csr_matrix((data, (rows, cols)), shape=(ng, len(all_keys[key])),
                                        dtype=np.int64)
                self._fit_keys = all_keys
                self._ngx = ng

            elif self._method_calling == 3:
                # A feature matrix for all levels
                M = dict()

                for (key, d) in filterfalse(lambda a: len(a[1]) == 0,
                                            iteritems(data)):
                    indexes, data = zip(*iteritems(d))
                    rows, cols = zip(*indexes)
                    M[key] = csr_matrix((data, (rows, cols)),
                                        shape=(ng, len(all_keys[key]) + len(self._fit_keys[key])),
                                        dtype=np.int64)

                self._ngy = ng

            return M

    def transform(self, X, y=None):
        """Calculate the kernel matrix, between given and fitted dataset.

        Parameters
        ----------
        X : iterable
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given graph
            format).

        y : Object, default=None
            Ignored argument, added for the pipeline.

        Returns
        -------
        K : numpy array, shape = [n_targets, n_input_graphs]
            corresponding to the kernel matrix, a calculation between
            all pairs of graphs between target an features

        """
        self._method_calling = 3
        # Check is fit had been called
        check_is_fitted(self, ['X'])

        # Input validation and parsing
        if X is None:
            raise ValueError('transform input cannot be None')
        else:
            Y = self.parse_input(X)

        try:
            check_is_fitted(self, ['_X_level_norm_factor'])
        except NotFittedError:
            self._X_level_norm_factor = \
                {key: np.array(M.power(2).sum(-1))
                 for (key, M) in iteritems(self.X)}

        N = self._X_level_norm_factor
        S = np.zeros(shape=(self._ngy, self._ngx))
        for (key, Mp) in filterfalse(lambda x: x[0] not in self.X,
                                     iteritems(Y)):
            M = self.X[key]
            K = M.dot(Mp.T[:M.shape[1]]).toarray().T
            S += np.nan_to_num(K / np.sqrt(np.outer(np.array(Mp.power(2).sum(-1)), N[key])))

        self._Y = Y
        self._is_transformed = True
        if self.normalize:
            S /= np.sqrt(np.outer(*self.diagonal()))
        return S

    def fit_transform(self, X):
        """Fit and transform, on the same dataset.

        Parameters
        ----------
        X : iterable
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given graph
            format). If None the kernel matrix is calculated upon fit data.
            The test samples.

        Returns
        -------
        K : numpy array, shape = [n_input_graphs, n_input_graphs]
            corresponding to the kernel matrix, a calculation between
            all pairs of graphs between target an features

        """
        self._method_calling = 2
        self.fit(X)

        S, N = np.zeros(shape=(self._ngx, self._ngx)), dict()
        for (key, M) in iteritems(self.X):
            K = M.dot(M.T).toarray()
            K_diag = K.diagonal()
            N[key] = K_diag
            S += np.nan_to_num(K / np.sqrt(np.outer(K_diag, K_diag)))

        self._X_level_norm_factor = N

        if self.normalize:
            return S / len(self.X)
        else:
            return S

    def diagonal(self):
        """Calculate the kernel matrix diagonal of the fitted data.

        Static. Added for completeness.

        Parameters
        ----------
        None.

        Returns
        -------
        X_diag : int
            Always equal with r*d.

        Y_diag : int
            Always equal with r*d.

        """
        # constant based on normalization of krd
        check_is_fitted(self, ['X'])
        try:
            check_is_fitted(self, ['_X_diag'])
        except NotFittedError:
            # Calculate diagonal of X
            self._X_diag = len(self.X)

        try:
            check_is_fitted(self, ['_Y'])
            return self._X_diag, len(self._Y)
        except NotFittedError:
            return self._X_diag

    def _hash_neighborhoods(self, vertices, edges, Lv, Le, N, D_pair):
        """Hash all neighborhoods and all root nodes.

        Parameters
        ----------
        vertices : set
            The graph vertices.

        edges : set
            The set of edges

        N : dict
            Neighborhoods that map levels (int) to dictionaries of root node
            symbols (keys) to list of vertex symbols, which correspond to the
            neighbors, that belong to this neighborhood.

        D_pairs : dict
            A dictionary that maps edges (tuple pairs of vertex symbols) to
            element distances (int - as produced from a BFS traversal).

        Returns
        -------
        H : dict
            The hashed neighborhoods as a 2-level dict from radious,
            vertex to the hashed values.

        """
        H, sel = dict(), sorted(list(edges))
        for v in vertices:
            re, lv, le = sel, Lv, Le
            for radius in range(self.r, -1, -1):
                sub_vertices = sorted(N[radius][v])
                re = {(i, j) for (i, j) in re
                      if i in sub_vertices and j in sub_vertices}
                lv = {v: lv[v] for v in sub_vertices}
                le = {e: le[e] for e in edges}
                H[radius, v] = hash_graph(D_pair, sub_vertices, re, lv, le)
        return H


def hash_graph(D, vertices, edges, glv, gle):
    """Make labels for hashing according to the proposed method.

    Produces the graph hash needed for fast comparison.

    Parameters
    ----------
    D_pairs : dict
        A dictionary that maps edges (tuple pairs of vertex symbols) to
        element distances (int - as produced from a BFS traversal).

    vertices : set
        A set of vertices.

    edges : set
        A set of edges.

    glv : dict
        Labels for vertices of the graph.

    gle : dict
        Labels for edges of the graph.

    Returns
    -------
    hash : int.
        The hash value for the given graph.

    """
    encoding = ""

    # Make labels for vertices
    Lv = dict()
    for i in vertices:
        label = "|".join(sorted([str(D[(i, j)]) + ',' + str(glv[j])
                                 for j in vertices if (i, j) in D]))
        encoding += label + "."
        Lv[i] = label

    encoding = encoding[:-1]+":"

    # Expand to labels for edges
    for (i, j) in edges:
        encoding += Lv[i] + ',' + Lv[j] + ',' + str(gle[(i, j)]) + "_"

    # Arash Partov hashing, as in the original
    # implementation of NSPK.
    return APHash(encoding)



"""The weisfeiler lehman kernel :cite:`shervashidze2011weisfeiler`."""
# Author: Ioannis Siglidis <y.siglidis@gmail.com>
# License: BSD 3 clause
import warnings

import numpy as np
import joblib

from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from grakel.graph import Graph
from grakel.kernels import Kernel
from grakel.kernels.vertex_histogram import VertexHistogram

# Python 2/3 cross-compatibility import
from six import iteritems
from six import itervalues
from six.moves.collections_abc import Iterable


class WeisfeilerLehman(Kernel):
    """Compute the Weisfeiler Lehman Kernel.

     See :cite:`shervashidze2011weisfeiler`.

    Parameters
    ----------
    n_iter : int, default=5
        The number of iterations.

    base_graph_kernel : `grakel.kernels.Kernel` or tuple, default=None
        If tuple it must consist of a valid kernel object and a
        dictionary of parameters. General parameters concerning
        normalization, concurrency, .. will be ignored, and the
        ones of given on `__init__` will be passed in case it is needed.
        Default `base_graph_kernel` is `VertexHistogram`.

    Attributes
    ----------
    X : dict
     Holds a dictionary of fitted subkernel modules for all levels.

    _nx : number
        Holds the number of inputs.

    _n_iter : int
        Holds the number, of iterations.

    _base_graph_kernel : function
        A void function that initializes a base kernel object.

    _inv_labels : dict
        An inverse dictionary, used for relabeling on each iteration.

    """

    _graph_format = "dictionary"

    def __init__(self, n_jobs=None, verbose=False,
                 normalize=False, n_iter=5, base_graph_kernel=VertexHistogram):
        """Initialise a `weisfeiler_lehman` kernel."""
        super(WeisfeilerLehman, self).__init__(
            n_jobs=n_jobs, verbose=verbose, normalize=normalize)

        self.n_iter = n_iter
        self.base_graph_kernel = base_graph_kernel
        self._initialized.update({"n_iter": False, "base_graph_kernel": False})
        self._base_graph_kernel = None

    def initialize(self):
        """Initialize all transformer arguments, needing initialization."""
        super(WeisfeilerLehman, self).initialize()
        if not self._initialized["base_graph_kernel"]:
            base_graph_kernel = self.base_graph_kernel
            if base_graph_kernel is None:
                base_graph_kernel, params = VertexHistogram, dict()
            elif type(base_graph_kernel) is type and issubclass(base_graph_kernel, Kernel):
                params = dict()
            else:
                try:
                    base_graph_kernel, params = base_graph_kernel
                except Exception:
                    raise TypeError('Base kernel was not formulated in '
                                    'the correct way. '
                                    'Check documentation.')

                if not (type(base_graph_kernel) is type and
                        issubclass(base_graph_kernel, Kernel)):
                    raise TypeError('The first argument must be a valid '
                                    'grakel.kernel.kernel Object')
                if type(params) is not dict:
                    raise ValueError('If the second argument of base '
                                     'kernel exists, it must be a diction'
                                     'ary between parameters names and '
                                     'values')
                params.pop("normalize", None)

            params["normalize"] = False
            params["verbose"] = self.verbose
            params["n_jobs"] = None
            self._base_graph_kernel = base_graph_kernel
            self._params = params
            self._initialized["base_graph_kernel"] = True

        if not self._initialized["n_iter"]:
            if type(self.n_iter) is not int or self.n_iter <= 0:
                raise TypeError("'n_iter' must be a positive integer")
            self._n_iter = self.n_iter + 1
            self._initialized["n_iter"] = True

    def parse_input(self, X):
        """Parse input for weisfeiler lehman.

        Parameters
        ----------
        X : iterable
            For the input to pass the test, we must have:
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that correspond to the given
            graph format). A valid input also consists of graph type objects.

        Returns
        -------
        base_graph_kernel : object
        Returns base_graph_kernel.

        """
        if self._method_calling not in [1, 2]:
            raise ValueError('method call must be called either from fit ' +
                             'or fit-transform')
        elif hasattr(self, '_X_diag'):
            # Clean _X_diag value
            delattr(self, '_X_diag')

        # Input validation and parsing
        if not isinstance(X, Iterable):
            raise TypeError('input must be an iterable\n')
        else:
            nx = 0
            Gs_ed, L, distinct_values, extras = dict(), dict(), set(), dict()
            for (idx, x) in enumerate(iter(X)):
                is_iter = isinstance(x, Iterable)
                if is_iter:
                    x = list(x)
                if is_iter and (len(x) == 0 or len(x) >= 2):
                    if len(x) == 0:
                        warnings.warn('Ignoring empty element on index: '
                                      + str(idx))
                        continue
                    else:
                        if len(x) > 2:
                            extra = tuple()
                            if len(x) > 3:
                                extra = tuple(x[3:])
                            x = Graph(x[0], x[1], x[2], graph_format=self._graph_format)
                            extra = (x.get_labels(purpose=self._graph_format,
                                                  label_type="edge", return_none=True), ) + extra
                        else:
                            x = Graph(x[0], x[1], {}, graph_format=self._graph_format)
                            extra = tuple()

                elif type(x) is Graph:
                    x.desired_format(self._graph_format)
                    el = x.get_labels(purpose=self._graph_format, label_type="edge", return_none=True)
                    if el is None:
                        extra = tuple()
                    else:
                        extra = (el, )

                else:
                    raise TypeError('each element of X must be either a ' +
                                    'graph object or a list with at least ' +
                                    'a graph like object and node labels ' +
                                    'dict \n')
                Gs_ed[nx] = x.get_edge_dictionary()
                L[nx] = x.get_labels(purpose="dictionary")
                extras[nx] = extra
                distinct_values |= set(itervalues(L[nx]))
                nx += 1
            if nx == 0:
                raise ValueError('parsed input is empty')

        # Save the number of "fitted" graphs.
        self._nx = nx

        # get all the distinct values of current labels
        WL_labels_inverse = dict()

        # assign a number to each label
        label_count = 0
        for dv in sorted(list(distinct_values)):
            WL_labels_inverse[dv] = label_count
            label_count += 1

        # Initalize an inverse dictionary of labels for all iterations
        self._inv_labels = dict()
        self._inv_labels[0] = WL_labels_inverse

        def generate_graphs(label_count, WL_labels_inverse):
            new_graphs = list()
            for j in range(nx):
                new_labels = dict()
                for k in L[j].keys():
                    new_labels[k] = WL_labels_inverse[L[j][k]]
                L[j] = new_labels
                # add new labels
                new_graphs.append((Gs_ed[j], new_labels) + extras[j])
            yield new_graphs

            for i in range(1, self._n_iter):
                label_set, WL_labels_inverse, L_temp = set(), dict(), dict()
                for j in range(nx):
                    # Find unique labels and sort
                    # them for both graphs
                    # Keep for each node the temporary
                    L_temp[j] = dict()
                    for v in Gs_ed[j].keys():
                        credential = str(L[j][v]) + "," + \
                            str(sorted([L[j][n] for n in Gs_ed[j][v].keys()]))
                        L_temp[j][v] = credential
                        label_set.add(credential)

                label_list = sorted(list(label_set))
                for dv in label_list:
                    WL_labels_inverse[dv] = label_count
                    label_count += 1

                # Recalculate labels
                new_graphs = list()
                for j in range(nx):
                    new_labels = dict()
                    for k in L_temp[j].keys():
                        new_labels[k] = WL_labels_inverse[L_temp[j][k]]
                    L[j] = new_labels
                    # relabel
                    new_graphs.append((Gs_ed[j], new_labels) + extras[j])
                self._inv_labels[i] = WL_labels_inverse
                yield new_graphs

        base_graph_kernel = {i: self._base_graph_kernel(**self._params) for i in range(self._n_iter)}
        if self._parallel is None:
            if self._method_calling == 1:
                for (i, g) in enumerate(generate_graphs(label_count, WL_labels_inverse)):
                    base_graph_kernel[i].fit(g)
            elif self._method_calling == 2:
                graphs = generate_graphs(label_count, WL_labels_inverse)
                values = [
                    base_graph_kernel[i].fit_transform(g) for (i, g) in enumerate(graphs)
                ]
                K = np.sum(values, axis=0)

        else:
            if self._method_calling == 1:
                self._parallel(joblib.delayed(efit)(base_graph_kernel[i], g)
                               for (i, g) in enumerate(generate_graphs(label_count, WL_labels_inverse)))
            elif self._method_calling == 2:
                K = np.sum(self._parallel(joblib.delayed(efit_transform)(base_graph_kernel[i], g)
                           for (i, g) in enumerate(generate_graphs(label_count, WL_labels_inverse))),
                           axis=0)

        if self._method_calling == 1:
            return base_graph_kernel
        elif self._method_calling == 2:
            return K, base_graph_kernel

    def fit_transform(self, X, y=None):
        """Fit and transform, on the same dataset.

        Parameters
        ----------
        X : iterable
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given graph
            format). If None the kernel matrix is calculated upon fit data.
            The test samples.

        y : Object, default=None
            Ignored argument, added for the pipeline.

        Returns
        -------
        K : numpy array, shape = [n_targets, n_input_graphs]
            corresponding to the kernel matrix, a calculation between
            all pairs of graphs between target an features

        """
        self._method_calling = 2
        self._is_transformed = False
        self.initialize()
        if X is None:
            raise ValueError('transform input cannot be None')
        else:
            km, self.X = self.parse_input(X)

        self._X_diag = np.diagonal(km)
        if self.normalize:
            old_settings = np.seterr(divide='ignore')
            km = np.nan_to_num(np.divide(km, np.sqrt(np.outer(self._X_diag, self._X_diag))))
            np.seterr(**old_settings)
        return km

    def transform(self, X):
        """Calculate the kernel matrix, between given and fitted dataset.

        Parameters
        ----------
        X : iterable
            Each element must be an iterable with at most three features and at
            least one. The first that is obligatory is a valid graph structure
            (adjacency matrix or edge_dictionary) while the second is
            node_labels and the third edge_labels (that fitting the given graph
            format). If None the kernel matrix is calculated upon fit data.
            The test samples.

        Returns
        -------
        K : numpy array, shape = [n_targets, n_input_graphs]
            corresponding to the kernel matrix, a calculation between
            all pairs of graphs between target an features

        """
        self._method_calling = 3
        # Check is fit had been called
        check_is_fitted(self, ['X', '_nx', '_inv_labels'])

        # Input validation and parsing
        if X is None:
            raise ValueError('transform input cannot be None')
        else:
            if not isinstance(X, Iterable):
                raise ValueError('input must be an iterable\n')
            else:
                nx = 0
                distinct_values = set()
                Gs_ed, L = dict(), dict()
                for (i, x) in enumerate(iter(X)):
                    is_iter = isinstance(x, Iterable)
                    if is_iter:
                        x = list(x)
                    if is_iter and len(x) in [0, 2, 3]:
                        if len(x) == 0:
                            warnings.warn('Ignoring empty element on index: '
                                          + str(i))
                            continue

                        elif len(x) in [2, 3]:
                            x = Graph(x[0], x[1], {}, self._graph_format)
                    elif type(x) is Graph:
                        x.desired_format("dictionary")
                    else:
                        raise ValueError('each element of X must have at ' +
                                         'least one and at most 3 elements\n')
                    Gs_ed[nx] = x.get_edge_dictionary()
                    L[nx] = x.get_labels(purpose="dictionary")

                    # Hold all the distinct values
                    distinct_values |= set(
                        v for v in itervalues(L[nx])
                        if v not in self._inv_labels[0])
                    nx += 1
                if nx == 0:
                    raise ValueError('parsed input is empty')

        nl = len(self._inv_labels[0])
        WL_labels_inverse = {dv: idx for (idx, dv) in
                             enumerate(sorted(list(distinct_values)), nl)}

        def generate_graphs(WL_labels_inverse, nl):
            # calculate the kernel matrix for the 0 iteration
            new_graphs = list()
            for j in range(nx):
                new_labels = dict()
                for (k, v) in iteritems(L[j]):
                    if v in self._inv_labels[0]:
                        new_labels[k] = self._inv_labels[0][v]
                    else:
                        new_labels[k] = WL_labels_inverse[v]
                L[j] = new_labels
                # produce the new graphs
                new_graphs.append([Gs_ed[j], new_labels])
            yield new_graphs

            for i in range(1, self._n_iter):
                new_graphs = list()
                L_temp, label_set = dict(), set()
                nl += len(self._inv_labels[i])
                for j in range(nx):
                    # Find unique labels and sort them for both graphs
                    # Keep for each node the temporary
                    L_temp[j] = dict()
                    for v in Gs_ed[j].keys():
                        credential = str(L[j][v]) + "," + \
                            str(sorted([L[j][n] for n in Gs_ed[j][v].keys()]))
                        L_temp[j][v] = credential
                        if credential not in self._inv_labels[i]:
                            label_set.add(credential)

                # Calculate the new label_set
                WL_labels_inverse = dict()
                if len(label_set) > 0:
                    for dv in sorted(list(label_set)):
                        idx = len(WL_labels_inverse) + nl
                        WL_labels_inverse[dv] = idx

                # Recalculate labels
                new_graphs = list()
                for j in range(nx):
                    new_labels = dict()
                    for (k, v) in iteritems(L_temp[j]):
                        if v in self._inv_labels[i]:
                            new_labels[k] = self._inv_labels[i][v]
                        else:
                            new_labels[k] = WL_labels_inverse[v]
                    L[j] = new_labels
                    # Create the new graphs with the new labels.
                    new_graphs.append([Gs_ed[j], new_labels])
                yield new_graphs

        if self._parallel is None:
            # Calculate the kernel matrix without parallelization
            graphs = generate_graphs(WL_labels_inverse, nl)
            values = [self.X[i].transform(g) for (i, g) in enumerate(graphs)]
            K = np.sum(values, axis=0)
        else:
            # Calculate the kernel marix with parallelization
            K = np.sum(self._parallel(joblib.delayed(etransform)(self.X[i], g) for (i, g)
                       in enumerate(generate_graphs(WL_labels_inverse, nl))), axis=0)

        self._is_transformed = True
        if self.normalize:
            X_diag, Y_diag = self.diagonal()
            old_settings = np.seterr(divide='ignore')
            K = np.nan_to_num(np.divide(K, np.sqrt(np.outer(Y_diag, X_diag))))
            np.seterr(**old_settings)

        return K

    def diagonal(self):
        """Calculate the kernel matrix diagonal for fitted data.

        A funtion called on transform on a seperate dataset to apply
        normalization on the exterior.

        Parameters
        ----------
        None.

        Returns
        -------
        X_diag : np.array
            The diagonal of the kernel matrix, of the fitted data.
            This consists of kernel calculation for each element with itself.

        Y_diag : np.array
            The diagonal of the kernel matrix, of the transformed data.
            This consists of kernel calculation for each element with itself.

        """
        # Check if fit had been called
        check_is_fitted(self, ['X'])
        try:
            check_is_fitted(self, ['_X_diag'])
            if self._is_transformed:
                Y_diag = self.X[0].diagonal()[1]
                for i in range(1, self._n_iter):
                    Y_diag += self.X[i].diagonal()[1]
        except NotFittedError:
            # Calculate diagonal of X
            if self._is_transformed:
                X_diag, Y_diag = self.X[0].diagonal()
                # X_diag is considered a mutable and should not affect the kernel matrix itself.
                X_diag.flags.writeable = True
                for i in range(1, self._n_iter):
                    x, y = self.X[i].diagonal()
                    X_diag += x
                    Y_diag += y
                self._X_diag = X_diag
            else:
                # case sub kernel is only fitted
                X_diag = self.X[0].diagonal()
                # X_diag is considered a mutable and should not affect the kernel matrix itself.
                X_diag.flags.writeable = True
                for i in range(1, self._n_iter):
                    x = self.X[i].diagonal()
                    X_diag += x
                self._X_diag = X_diag

        if self._is_transformed:
            return self._X_diag, Y_diag
        else:
            return self._X_diag


def efit(object, data):
    """Fit an object on data."""
    object.fit(data)


def efit_transform(object, data):
    """Fit-Transform an object on data."""
    return object.fit_transform(data)


def etransform(object, data):
    """Transform an object on data."""
    return object.transform(data)
